from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IDefinition,
    TokenInfo,
    IKeywordDefinition,
    IKeywordCollector,
    IVariablesCollector,
    IKeywordFound,
    AbstractKeywordCollector,
    IRobotToken,
    IVariableFound,
    AbstractVariablesCollector,
)
from robocorp_ls_core.protocols import check_implements, IDocumentSelection
from typing import Optional, Sequence, List
from robocorp_ls_core.lsp import RangeTypedDict, MarkupContentTypedDict, MarkupKind
from robocorp_ls_core.basic import isinstance_name


class _DefinitionFromKeyword(object):
    def __init__(self, keyword_found: IKeywordFound):
        self.keyword_found = keyword_found
        self.keyword_name = keyword_found.keyword_name
        self.source = keyword_found.source

        self.lineno = keyword_found.lineno
        self.end_lineno = keyword_found.end_lineno
        self.col_offset = keyword_found.col_offset
        self.end_col_offset = keyword_found.end_col_offset

        self.scope_lineno: Optional[int] = keyword_found.scope_lineno
        self.scope_end_lineno: Optional[int] = keyword_found.scope_end_lineno
        self.scope_col_offset: Optional[int] = keyword_found.scope_col_offset
        self.scope_end_col_offset: Optional[int] = keyword_found.scope_end_col_offset

    def hover_docs(self) -> MarkupContentTypedDict:
        from robotframework_ls.html_to_markdown import escape

        keyword_name = escape(self.keyword_name)
        return {"kind": MarkupKind.Markdown, "value": f"Keyword: **{keyword_name}**"}

    def __str__(self):
        return "DefinitionFromKeyword[%s, %s:%s]" % (
            self.keyword_name,
            self.source,
            self.lineno,
        )

    __repr__ = __str__

    def __typecheckself__(self) -> None:
        _: IKeywordDefinition = check_implements(self)


class _DefinitionFromLibrary(object):
    def __init__(self, library_doc):
        """
        :param LibraryDoc library_info:
        """
        self.keyword_name = ""
        self.library_doc = library_doc
        self.source = library_doc.source
        # Note: line/offsets 0-based.
        self.lineno = 0
        self.end_lineno = 0
        self.col_offset = 0
        self.end_col_offset = 0

        self.scope_lineno: Optional[int] = None
        self.scope_end_lineno: Optional[int] = None
        self.scope_col_offset: Optional[int] = None
        self.scope_end_col_offset: Optional[int] = None

    def hover_docs(self) -> MarkupContentTypedDict:
        from robotframework_ls.html_to_markdown import escape
        from robotframework_ls.impl.robot_specbuilder import docs_and_format

        library_name = escape(self.library_doc.name)
        docs, doc_format = docs_and_format(self.library_doc)

        return {
            "kind": doc_format,
            "value": f"Library: **{library_name}**\n\n{docs}",
        }

    def __str__(self):
        return "DefinitionFromLibrary[%s]" % (self.source,)

    __repr__ = __str__

    def __typecheckself__(self) -> None:
        _: IDefinition = check_implements(self)


class _DefinitionFromResource(object):
    def __init__(self, resource_doc):
        """
        :param RobotDocument resource_doc:
        """
        from robocorp_ls_core import uris

        self.keyword_name = ""
        self.resource_doc = resource_doc
        self.source = uris.to_fs_path(resource_doc.uri)
        # Note: line/offsets 0-based.
        self.lineno = 0
        self.end_lineno = 0
        self.col_offset = 0
        self.end_col_offset = 0

        self.scope_lineno: Optional[int] = None
        self.scope_end_lineno: Optional[int] = None
        self.scope_col_offset: Optional[int] = None
        self.scope_end_col_offset: Optional[int] = None

    def hover_docs(self) -> MarkupContentTypedDict:
        from robotframework_ls.impl import ast_utils
        from robotframework_ls.html_to_markdown import escape
        import os

        ast = self.resource_doc.get_ast()
        documentation = ast_utils.get_documentation_as_markdown(ast)
        if documentation:
            documentation = "\n\n" + documentation

        name = escape(os.path.splitext(os.path.basename(self.source))[0])

        return {
            "kind": MarkupKind.Markdown,
            "value": f"Resource: {name}{documentation}",
        }

    def __str__(self):
        return "DefinitionFromResource[%s]" % (self.source,)

    __repr__ = __str__

    def __typecheckself__(self) -> None:
        _: IDefinition = check_implements(self)


class _DefinitionFromVariableImport(object):
    def __init__(self, variables_doc):
        """
        :param RobotDocument variables_doc:
        """
        from robocorp_ls_core import uris

        self.keyword_name = ""
        self.variables_doc = variables_doc
        self.source = uris.to_fs_path(variables_doc.uri)
        # Note: line/offsets 0-based.
        self.lineno = 0
        self.end_lineno = 0
        self.col_offset = 0
        self.end_col_offset = 0

        self.scope_lineno: Optional[int] = None
        self.scope_end_lineno: Optional[int] = None
        self.scope_col_offset: Optional[int] = None
        self.scope_end_col_offset: Optional[int] = None

    def hover_docs(self) -> MarkupContentTypedDict:
        from robotframework_ls.html_to_markdown import escape
        import os

        name = escape(os.path.splitext(os.path.basename(self.source))[0])

        return {"kind": MarkupKind.Markdown, "value": f"Variable Import: {name}"}

    def __str__(self):
        return "DefinitionFromVariableImport[%s]" % (self.source,)

    __repr__ = __str__

    def __typecheckself__(self) -> None:
        _: IDefinition = check_implements(self)


class _DefinitionFromVariable(object):
    def __init__(self, variable_found):
        """
        :param IVariableFound variable_found:
        """
        self.variable_found = variable_found

        self.keyword_name = ""
        self.source = variable_found.source
        self.lineno = variable_found.lineno
        self.end_lineno = variable_found.end_lineno
        self.col_offset = variable_found.col_offset
        self.end_col_offset = variable_found.end_col_offset

        self.scope_lineno: Optional[int] = None
        self.scope_end_lineno: Optional[int] = None
        self.scope_col_offset: Optional[int] = None
        self.scope_end_col_offset: Optional[int] = None

    def hover_docs(self) -> MarkupContentTypedDict:
        from robotframework_ls.html_to_markdown import escape

        variable_found = self.variable_found
        variable_name = "".join(
            (
                variable_found.variable_kind,
                ": ",
                "**",
                escape(variable_found.variable_name.strip()),
                "**",
            )
        )

        variable_value = ""
        if variable_found.variable_value:
            variable_value = "".join(
                (
                    "\n\nValue: ",
                    escape(variable_found.variable_value.strip()),
                )
            )

        return {
            "kind": MarkupKind.Markdown,
            "value": (f"{variable_name}{variable_value}"),
        }

    def __str__(self):
        return "_DefinitionFromVariable(%s[%s, %s:%s])" % (
            self.variable_found.__class__.__name__,
            self.variable_found.variable_name,
            self.source,
            self.lineno,
        )

    __repr__ = __str__

    def __typecheckself__(self) -> None:
        _: IDefinition = check_implements(self)


class _FindDefinitionKeywordCollector(AbstractKeywordCollector):
    def __init__(self, match_name):
        from robotframework_ls.impl.string_matcher import RobotStringMatcher
        from robotframework_ls.impl.string_matcher import (
            build_matchers_with_resource_or_library_scope,
        )

        self.match_name = match_name
        self.matches = []

        self._matcher = RobotStringMatcher(match_name)
        self._scope_matchers = build_matchers_with_resource_or_library_scope(match_name)

    def accepts(self, keyword_name):
        return True

    def on_keyword(self, keyword_found):
        if self._matcher.is_keyword_name_match(keyword_found.keyword_name):
            definition = _DefinitionFromKeyword(keyword_found)
            self.matches.append(definition)
            return

        for matcher in self._scope_matchers:
            if matcher.is_keyword_match(keyword_found):
                definition = _DefinitionFromKeyword(keyword_found)
                self.matches.append(definition)
                return

    def __typecheckself__(self) -> None:
        _: IKeywordCollector = check_implements(self)


class _FindDefinitionVariablesCollector(AbstractVariablesCollector):
    def __init__(
        self, sel: IDocumentSelection, token: IRobotToken, robot_string_matcher
    ):
        self.matches: List[IDefinition] = []
        self.sel = sel
        self.token = token
        self.matcher = robot_string_matcher

    def accepts(self, variable_name: str) -> bool:
        return self.matcher.is_variable_name_match(variable_name)

    def on_variable(self, variable_found: IVariableFound):
        definition = _DefinitionFromVariable(variable_found)
        self.matches.append(definition)

    def __typecheckself__(self) -> None:
        _: IVariablesCollector = check_implements(self)


def find_keyword_definition(
    completion_context: ICompletionContext, token_info: TokenInfo
) -> Optional[Sequence[IKeywordDefinition]]:
    """
    Find a definition only considering Keywords.

    The token info must be already computed and must match the completion
    context location.
    """
    from robotframework_ls.impl.collect_keywords import collect_keywords
    from robotframework_ls.impl import ast_utils

    token = ast_utils.get_keyword_name_token(
        token_info.stack, token_info.node, token_info.token
    )
    if token is None:
        if token_info.token.type == token_info.token.KEYWORD_NAME:
            if isinstance_name(token_info.node, "KeywordName"):
                token = token_info.token

    if token is not None:
        collector = _FindDefinitionKeywordCollector(token.value)
        collect_keywords(completion_context, collector)
        return collector.matches
    return None


def find_definition(completion_context: ICompletionContext) -> Sequence[IDefinition]:
    definition_info = find_definition_extended(completion_context)
    if definition_info is None:
        return []
    return definition_info.definitions


class _DefinitionInfo:
    def __init__(
        self,
        definitions: Sequence[IDefinition],
        origin_selection_range: Optional[RangeTypedDict] = None,
    ):
        self.definitions = definitions
        self.origin_selection_range = origin_selection_range


def find_definition_extended(
    completion_context: ICompletionContext,
) -> Optional[_DefinitionInfo]:
    """
    :note:
        Definitions may be found even if a given source file no longer exists
        at this place (callers are responsible for validating entries).
    """
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robotframework_ls.impl.variable_completions import collect_variables
    from robotframework_ls.impl.libspec_manager import LibspecManager

    token_info = completion_context.get_current_token()

    if token_info is not None:
        matches = find_keyword_definition(completion_context, token_info)
        if matches is not None:
            return _DefinitionInfo(
                matches, ast_utils.create_range_from_token(token_info.token)
            )

        token = ast_utils.get_library_import_name_token(
            token_info.node, token_info.token
        )
        if token is not None:
            libspec_manager: LibspecManager = (
                completion_context.workspace.libspec_manager
            )
            completion_context.check_cancelled()
            library_doc = libspec_manager.get_library_doc_or_error(
                completion_context.token_value_resolving_variables(token),
                create=True,
                completion_context=completion_context,
                args=ast_utils.get_library_arguments_serialized(token_info.node),
            ).library_doc
            if library_doc is not None:
                definition = _DefinitionFromLibrary(library_doc)
                return _DefinitionInfo(
                    [definition], ast_utils.create_range_from_token(token)
                )

        token = ast_utils.get_resource_import_name_token(
            token_info.node, token_info.token
        )
        if token is not None:
            completion_context.check_cancelled()
            resource_import_as_doc = completion_context.get_resource_import_as_doc(
                token_info.node
            )
            if resource_import_as_doc is not None:
                return _DefinitionInfo(
                    [
                        _DefinitionFromResource(resource_import_as_doc),
                    ],
                    ast_utils.create_range_from_token(token),
                )

        token = ast_utils.get_variables_import_name_token(
            token_info.node, token_info.token
        )
        if token is not None:
            completion_context.check_cancelled()
            variable_import_as_doc = completion_context.get_variable_import_as_doc(
                token_info.node
            )
            if variable_import_as_doc is not None:
                return _DefinitionInfo(
                    [_DefinitionFromVariableImport(variable_import_as_doc)],
                    ast_utils.create_range_from_token(token),
                )

    var_token_info = completion_context.get_current_variable()
    if var_token_info is not None:
        token = var_token_info.token
        value = token.value
        collector = _FindDefinitionVariablesCollector(
            completion_context.sel, token, RobotStringMatcher(value)
        )
        collect_variables(completion_context, collector)
        return _DefinitionInfo(
            collector.matches, ast_utils.create_range_from_token(token)
        )

    return None



def foo():
    return 1  # breakpoint


def main():
    import linecache
    import time

    code = 'foo()'

    # Intermediate <string> stack frame will have no source.
    eval(code)

    co_filename = '<something>'
    co = compile(code, co_filename, 'exec')

    # Set up the source in linecache. Python doesn't have a public API for
    # this, so we have to hack around it, similar to what IPython does.
    linecache.cache[co_filename] = (
        len(code),
        time.time(),
        [line + '\n' for line in code.splitlines()],
        co_filename,
    )

    # Intermediate <something> stack frame will have source.
    eval(co)


if __name__ == '__main__':
    main()
    print('TEST SUCEEDED!')



'''
Created on Aug 22, 2011

@author: hussain.bohra
@author: fabioz
'''

import os
import sys
import unittest

#=======================================================================================================================
# Test
#=======================================================================================================================
class Test(unittest.TestCase):
    """Test cases to validate custom property implementation in pydevd
    """

    def setUp(self, nused=None):
        self.tempdir = os.path.join(os.path.dirname(os.path.dirname(__file__)))
        sys.path.insert(0, self.tempdir)
        from _pydevd_bundle import pydevd_traceproperty
        self.old = pydevd_traceproperty.replace_builtin_property()


    def tearDown(self, unused=None):
        from _pydevd_bundle import pydevd_traceproperty
        pydevd_traceproperty.replace_builtin_property(self.old)
        sys.path.remove(self.tempdir)


    def test_property(self):
        """Test case to validate custom property
        """

        from _pydevd_bundle import pydevd_traceproperty
        class TestProperty(object):

            def __init__(self):
                self._get = 0
                self._set = 0
                self._del = 0

            def get_name(self):
                self._get += 1
                return self.__name

            def set_name(self, value):
                self._set += 1
                self.__name = value

            def del_name(self):
                self._del += 1
                del self.__name
            name = property(get_name, set_name, del_name, "name's docstring")
            self.assertEqual(name.__class__, pydevd_traceproperty.DebugProperty)

        testObj = TestProperty()
        self._check(testObj)


    def test_property2(self):
        """Test case to validate custom property
        """

        class TestProperty(object):

            def __init__(self):
                self._get = 0
                self._set = 0
                self._del = 0

            def name(self):
                self._get += 1
                return self.__name
            name = property(name)

            def set_name(self, value):
                self._set += 1
                self.__name = value
            name.setter(set_name)

            def del_name(self):
                self._del += 1
                del self.__name
            name.deleter(del_name)

        testObj = TestProperty()
        self._check(testObj)


    def test_property3(self):
        """Test case to validate custom property
        """

        class TestProperty(object):

            def __init__(self):
                self._name = 'foo'

            def name(self):
                return self._name
            name = property(name)

        testObj = TestProperty()
        self.assertRaises(AttributeError, setattr, testObj, 'name', 'bar')
        self.assertRaises(AttributeError, delattr, testObj, 'name')


    def _check(self, testObj):
        testObj.name = "Custom"
        self.assertEqual(1, testObj._set)

        self.assertEqual(testObj.name, "Custom")
        self.assertEqual(1, testObj._get)

        self.assertTrue(hasattr(testObj, 'name'))
        del testObj.name
        self.assertEqual(1, testObj._del)

        self.assertTrue(not hasattr(testObj, 'name'))
        testObj.name = "Custom2"
        self.assertEqual(testObj.name, "Custom2")




import unittest

class SetUpClassTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        raise ValueError("This is an INTENTIONAL value error in setUpClass.")

    def test_blank(self):
        pass


if __name__ == '__main__':
    unittest.main()



# Note: this module should be self-contained to run tests (as it relies on `threading` not being
# imported and having no other threads running).


def wait_for_condition(condition, msg=None, timeout=5, sleep=.05):
    import time
    curtime = time.time()
    while True:
        if condition():
            break
        if time.time() - curtime > timeout:
            error_msg = 'Condition not reached in %s seconds' % (timeout,)
            if msg is not None:
                error_msg += '\n'
                if callable(msg):
                    error_msg += msg()
                else:
                    error_msg += str(msg)

            raise AssertionError('Timeout: %s' % (error_msg,))
        time.sleep(sleep)


def check_main_thread_id_simple():
    import attach_script
    import sys
    assert 'threading' not in sys.modules
    try:
        import thread
    except ImportError:
        import _thread as thread

    main_thread_id, log_msg = attach_script.get_main_thread_id(None)
    assert main_thread_id == thread.get_ident(), 'Found: %s, Expected: %s' % (main_thread_id, thread.get_ident())
    assert not log_msg
    assert 'threading' not in sys.modules
    wait_for_condition(lambda: len(sys._current_frames()) == 1)


def check_main_thread_id_multiple_threads():
    import attach_script
    import sys
    import time
    assert 'threading' not in sys.modules
    try:
        import thread
    except ImportError:
        import _thread as thread

    lock = thread.allocate_lock()
    lock2 = thread.allocate_lock()

    def method():
        lock2.acquire()
        with lock:
            pass  # Will only finish when lock is released.

    with lock:
        thread.start_new_thread(method, ())
        while not lock2.locked():
            time.sleep(.1)

        wait_for_condition(lambda: len(sys._current_frames()) == 2)

        main_thread_id, log_msg = attach_script.get_main_thread_id(None)
        assert main_thread_id == thread.get_ident(), 'Found: %s, Expected: %s' % (main_thread_id, thread.get_ident())
        assert not log_msg
        assert 'threading' not in sys.modules
    wait_for_condition(lambda: len(sys._current_frames()) == 1)


def check_fix_main_thread_id_multiple_threads():
    import attach_script
    import sys
    import time
    assert 'threading' not in sys.modules
    try:
        import thread
    except ImportError:
        import _thread as thread

    lock = thread.allocate_lock()
    lock2 = thread.allocate_lock()

    def method():
        lock2.acquire()
        import threading  # Note: imported on wrong thread
        assert threading.current_thread().ident == thread.get_ident()
        assert threading.current_thread() is attach_script.get_main_thread_instance(threading)

        attach_script.fix_main_thread_id()

        assert threading.current_thread().ident == thread.get_ident()
        assert threading.current_thread() is not attach_script.get_main_thread_instance(threading)

        with lock:
            pass  # Will only finish when lock is released.

    with lock:
        thread.start_new_thread(method, ())
        while not lock2.locked():
            time.sleep(.1)

        wait_for_condition(lambda: len(sys._current_frames()) == 2)

        main_thread_id, log_msg = attach_script.get_main_thread_id(None)
        assert main_thread_id == thread.get_ident(), 'Found: %s, Expected: %s' % (main_thread_id, thread.get_ident())
        assert not log_msg
        assert 'threading' in sys.modules
        import threading
        assert threading.current_thread().ident == main_thread_id
    wait_for_condition(lambda: len(sys._current_frames()) == 1)


def check_win_threads():
    import sys
    if sys.platform != 'win32':
        return

    import attach_script
    import time
    assert 'threading' not in sys.modules
    try:
        import thread
    except ImportError:
        import _thread as thread
    from ctypes import windll, WINFUNCTYPE, c_uint32, c_void_p, c_size_t

    ThreadProc = WINFUNCTYPE(c_uint32, c_void_p)

    lock = thread.allocate_lock()
    lock2 = thread.allocate_lock()

    @ThreadProc
    def method(_):
        lock2.acquire()
        with lock:
            pass  # Will only finish when lock is released.
        return 0

    with lock:
        windll.kernel32.CreateThread(None, c_size_t(0), method, None, c_uint32(0), None)
        while not lock2.locked():
            time.sleep(.1)

        wait_for_condition(lambda: len(sys._current_frames()) == 2)

        main_thread_id, log_msg = attach_script.get_main_thread_id(None)
        assert main_thread_id == thread.get_ident(), 'Found: %s, Expected: %s' % (main_thread_id, thread.get_ident())
        assert not log_msg
        assert 'threading' not in sys.modules
    wait_for_condition(lambda: len(sys._current_frames()) == 1)


if __name__ == '__main__':
    check_main_thread_id_simple()
    check_main_thread_id_multiple_threads()
    check_win_threads()
    check_fix_main_thread_id_multiple_threads()  # Important: must be the last test checked!



"""
This script is meant to be called as:

python run_and_save_pid.py <file_to_write_pid> <executable> <executable_args, ...>

After the launch is made, <file_to_write_pid> should have the <pid>\n written.
"""
from typing import List, Optional
import subprocess
import sys


def main(file_to_write_pid: str, args: List[str]):
    """
    The first argument is where we should write the pid to. All the
    other arguments are used to make the launch.
    """
    p = subprocess.Popen(args, shell=sys.platform == "win32")

    with open(file_to_write_pid, "w") as stream:
        stream.write(str(p.pid))
        stream.write("\n")
        stream.flush()

    returncode = p.poll()
    while returncode is None:
        try:
            p.wait()
        except:
            pass
        returncode = p.poll()

    return returncode


if __name__ == "__main__":
    # Remove current script from args.
    args = sys.argv[1:]
    sys.exit(main(args[0], args[1:]))


def wait_for_pid_in_file(target_file: str, timeout: float = 10) -> int:
    import time
    from pathlib import Path

    initial_time = time.time()

    p = Path(target_file)

    def read_pid() -> Optional[int]:
        if p.exists():
            txt = p.read_text("utf-8")
            if txt.endswith("\n"):
                return int(txt.strip())
        return None

    failures_reading = 0
    while True:
        pid = None
        try:
            pid = read_pid()
        except:
            failures_reading += 1
            if failures_reading > 2:
                raise

        if pid is not None:
            return pid
        if timeout is not None and (time.time() - initial_time > timeout):
            error_msg = f"Unable to read pid in {target_file} in {timeout} seconds"
            raise TimeoutError(error_msg)

        time.sleep(1.0 / 15.0)



import threading
if __name__ == '__main__':
    import os
    import sys
    port = int(sys.argv[1])
    root_dirname = os.path.dirname(os.path.dirname(__file__))

    if root_dirname not in sys.path:
        sys.path.append(root_dirname)

    def method(i):
        import time
        wait = True
        while wait:
            time.sleep(.1)  # break here

    threads = [threading.Thread(target=method, args=(i,)) for i in range(2)]
    for t in threads:
        t.start()

    import pydevd
    assert pydevd.get_global_debugger() is None

    print('before pydevd.settrace')
    pydevd.settrace(port=port)
    print('after pydevd.settrace')

    for t in threads:
        t.join()

    print('TEST SUCEEDED!')



from tests_python.resource_path_translation import other


def call_this():
    print('break here')


def main():
    other.call_me_back1(call_this)
    print('TEST SUCEEDED!')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Simple GTK example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for gtk
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console
"""

if __name__ == '__main__':
    import pygtk
    pygtk.require('2.0')
    import gtk
    
    
    def hello_world(wigdet, data=None):
        print("Hello World")
    
    def delete_event(widget, event, data=None):
        return False
    
    def destroy(widget, data=None):
        gtk.main_quit()
    
    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    window.connect("delete_event", delete_event)
    window.connect("destroy", destroy)
    button = gtk.Button("Hello World")
    button.connect("clicked", hello_world, None)
    
    window.add(button)
    button.show()
    window.show()
    



import os.path


class CasesFixture(object):
    def __init__(self, copy_to_dir: str, original_resources_dir: str):
        """
        Upon initialization copies the `original_resources_dir` to
        `copy_to_dir`.

        So, for instance, we may copy the contents from

        /my/test/resource

        to

        /temp/pytest-101/folder with spaces/resource

        Subsequent requests to get the path will access it in the
        place we copied it to.

        Note: it should usually be bound to a session scope so that
        the copy isn't done at each call.
        """
        from robocorp_ls_core.copytree import copytree_dst_exists

        copytree_dst_exists(original_resources_dir, copy_to_dir)
        self.resources_dir = copy_to_dir
        assert os.path.exists(self.resources_dir)

    def get_path(self, resources_relative_path: str, must_exist=True) -> str:
        """
        Returns a path from the resources dir.
        """
        path = os.path.join(self.resources_dir, resources_relative_path)
        if must_exist:
            assert os.path.exists(path), "%s does not exist." % (path,)
        return path

    def copy_to(self, case: str, dest_dir: str):
        """
        Helper to copy a given path to a given directory.

        To be used if a given path should be within another structure or
        if its contents should be mutated.
        """
        import shutil

        src = self.get_path(case, must_exist=True)

        shutil.copytree(src, dest_dir)



"""
Robocop rules are internally grouped into checkers. Each checker can scan for multiple related issues
(like ``LengthChecker`` checks both for minimum and maximum length of a keyword). You can refer to
specific rule reported by checkers by its name or id (for example `too-long-keyword` or `0501`).

.. dropdown:: Checker groups IDs:

    Checkers are categorized into following groups:
     * 01: base
     * 02: documentation
     * 03: naming
     * 04: errors
     * 05: lengths
     * 06: tags
     * 07: comments
     * 08: duplications
     * 09: misc
     * 10: spacing
     * 11-50: not yet used: reserved for future internal checkers
     * 51-99: reserved for external checkers

Checker has two basic types:

- ``VisitorChecker`` uses Robot Framework parsing API and Python `ast` module for traversing Robot code as nodes,

- ``RawFileChecker`` simply reads Robot file as normal file and scans every line.

Every rule has a `unique id` made of 4 digits where first 2 are `checker id` while 2 latter are `rule id`.
`Unique id` as well as `rule name` can be used to refer to the rule (e.g. in include/exclude statements,
configurations etc.). You can optionally configure rule severity or other parameters.
"""
import inspect

try:
    from robot.api.parsing import ModelVisitor
except ImportError:
    from robot.parsing.model.visitor import ModelVisitor

from robot.utils import FileReader

from robocop.utils import modules_from_paths, modules_in_current_dir
from robocop.exceptions import RuleNotFoundError, RuleParamNotFoundError, RuleReportsNotFoundError


class BaseChecker:
    rules = None

    def __init__(self):
        self.disabled = False
        self.source = None
        self.lines = None
        self.issues = []
        self.rules = {}
        self.templated_suite = False

    def param(self, rule, param_name):
        try:
            return self.rules[rule].config[param_name].value
        except KeyError:
            if rule not in self.rules:
                raise RuleNotFoundError(rule, self) from None
            if param_name not in self.rules[rule].config:
                raise RuleParamNotFoundError(self.rules[rule], param_name, self) from None

    def report(
        self,
        rule,
        node=None,
        lineno=None,
        col=None,
        end_lineno=None,
        end_col=None,
        ext_disablers=None,
        **kwargs,
    ):
        if rule not in self.rules:
            raise ValueError(f"Missing definition for message with name {rule}")
        message = self.rules[rule].prepare_message(
            source=self.source,
            node=node,
            lineno=lineno,
            col=col,
            end_lineno=end_lineno,
            end_col=end_col,
            ext_disablers=ext_disablers,
            **kwargs,
        )
        if message.enabled:
            self.issues.append(message)


class VisitorChecker(BaseChecker, ModelVisitor):  # noqa
    type = "visitor_checker"

    def scan_file(self, ast_model, filename, in_memory_content, templated=False):
        self.issues = []
        self.source = filename
        self.templated_suite = templated
        if in_memory_content is not None:
            self.lines = in_memory_content.splitlines(keepends=True)
        else:
            self.lines = None
        self.visit_File(ast_model)
        return self.issues

    def visit_File(self, node):  # noqa
        """Perform generic ast visit on file node."""
        self.generic_visit(node)


class RawFileChecker(BaseChecker):  # noqa
    type = "rawfile_checker"

    def scan_file(self, ast_model, filename, in_memory_content, templated=False):
        self.issues = []
        self.source = filename
        self.templated_suite = templated
        if in_memory_content is not None:
            self.lines = in_memory_content.splitlines(keepends=True)
        else:
            self.lines = None
        self.parse_file()
        return self.issues

    def parse_file(self):
        """Read file line by line and for each call check_line method."""
        if self.lines is not None:
            for lineno, line in enumerate(self.lines):
                self.check_line(line, lineno + 1)
        else:
            with FileReader(self.source) as file_reader:
                for lineno, line in enumerate(file_reader.readlines()):
                    self.check_line(line, lineno + 1)

    def check_line(self, line, lineno):
        raise NotImplementedError


def init(linter):
    """For each module get `rules` dictionary and visitors. Instantiate each visitor and map it to the
    rule class instance using `reports` visitor attribute."""
    for module in get_modules(linter.config.ext_rules):
        classes = inspect.getmembers(module, inspect.isclass)
        module_rules = {rule.name: rule for rule in getattr(module, "rules", {}).values()}
        for checker in classes:
            if issubclass(checker[1], BaseChecker) and getattr(checker[1], "reports", False):
                checker_instance = checker[1]()
                for reported_rule in checker_instance.reports:
                    if reported_rule not in module_rules:
                        raise RuleReportsNotFoundError(reported_rule, checker_instance) from None
                    checker_instance.rules[reported_rule] = module_rules[reported_rule]
                linter.register_checker(checker_instance)


def get_modules(ext_rules):
    yield from modules_in_current_dir(__file__, __name__)
    yield from modules_from_paths(ext_rules)


def get_rules():
    for module in modules_in_current_dir(__file__, __name__):
        module_name = module.__name__.split(".")[-1]
        for rule in getattr(module, "rules", {}).values():
            yield module_name, rule



""" Auto detect robot file type (it can be resource, general or init) """
import ast
import os
from enum import Enum
from pathlib import Path

from robot.api import get_init_model, get_model, get_resource_model
from robot.errors import DataError
from robot.utils.robotpath import find_file


class FileType(Enum):
    """
    Enum holding type of Robot file.
    """

    RESOURCE = "resource"
    GENERAL = "general"
    INIT = "init"

    def get_parser(self):
        """return parser (method) for given model type"""
        return {
            "resource": get_resource_model,
            "general": get_model,
            "init": get_init_model,
        }[self.value]


class FileTypeChecker(ast.NodeVisitor):
    """
    Check if file contains import statements. If the import is in list of files to be scanned, update its type
    from GENERAL to RESOURCE.
    """

    def __init__(self, exec_dir):
        self.resource_files = set()
        self.exec_dir = exec_dir
        self.source = None

    def visit_ResourceImport(self, node):  # noqa
        """
        Check all imports in scanned file. If one of our scanned file is imported somewhere else
        it means this file is resource type
        """
        path_normalized = normalize_robot_path(node.name, Path(self.source).parent, self.exec_dir)
        try:
            path_normalized = find_file(path_normalized, self.source.parent, file_type="Resource")
        except DataError:
            pass
        else:
            path = Path(path_normalized)
            self.resource_files.add(path)


def normalize_robot_path(robot_path, curr_path, exec_path):
    normalized_path = str(robot_path).replace("${/}", os.path.sep)
    normalized_path = normalized_path.replace("${CURDIR}", str(curr_path))
    normalized_path = normalized_path.replace("${EXECDIR}", str(exec_path))
    return normalized_path



from robot.api.parsing import ModelTransformer, Tags, Token, DefaultTags, ForceTags

from robotidy.exceptions import InvalidParameterValueError


class NormalizeTags(ModelTransformer):
    """
    Normalize tag names by normalizing case and removing duplicates.

    Example usage:

        robotidy --transform NormalizeTags:case=lowercase test.robot

    Other supported cases: uppercase, title case. The default is lowercase.

    You can also run it to remove duplicates but preserve current case by setting ``normalize_case`` parameter to False:

        robotidy --transform NormalizeTags:normalize_case=False test.robot

    See https://robotidy.readthedocs.io/en/latest/transformers/NormalizeTags.html for more examples.
    """

    CASE_FUNCTIONS = {
        "lowercase": str.lower,
        "uppercase": str.upper,
        "titlecase": str.title,
    }

    def __init__(self, case: str = "lowercase", normalize_case: bool = True):
        self.case = case.lower()
        self.normalize_case = normalize_case
        try:
            self.case_function = self.CASE_FUNCTIONS[self.case]
        except KeyError:
            raise InvalidParameterValueError(
                self.__class__.__name__, "case", case, "Supported cases: lowercase, uppercase, titlecase."
            )

    def visit_Tags(self, node):  # noqa
        return self.normalize_tags(node, Tags, indent=True)

    def visit_DefaultTags(self, node):  # noqa
        return self.normalize_tags(node, DefaultTags)

    def visit_ForceTags(self, node):  # noqa
        return self.normalize_tags(node, ForceTags)

    def normalize_tags(self, node, tag_class, indent=False):
        tags = [tag.value for tag in node.data_tokens[1:]]
        if self.normalize_case:
            tags = self.convert_case(tags)
        tags = self.remove_duplicates(tags)
        comments = node.get_tokens(Token.COMMENT)
        if indent:
            tag_node = tag_class.from_params(
                tags,
                indent=self.formatting_config.separator,
                separator=self.formatting_config.separator,
            )
        else:
            tag_node = tag_class.from_params(tags, separator=self.formatting_config.separator)
        if comments:
            tag_node.tokens = tag_node.tokens[:-1] + tuple(self.join_tokens(comments)) + (tag_node.tokens[-1],)
        return tag_node

    def convert_case(self, tags):
        return [self.case_function(item) for item in tags]

    @staticmethod
    def remove_duplicates(tags):
        return list(dict.fromkeys(tags))

    def join_tokens(self, tokens):
        joined_tokens = []
        for token in tokens:
            joined_tokens.append(Token(Token.SEPARATOR, self.formatting_config.separator))
            joined_tokens.append(token)
        return joined_tokens



import sys
import weakref

def set_up():
    observable = Observable()
    observer = Observer()
    observable.add_observer(observer)
    return observable


class Observable(object):
    def __init__(self):
        self.observers = []
        
    def add_observer(self, observer):
        sys.stdout.write( 'observer %s\n' % (observer,))
        ref = weakref.ref(observer)
        self.observers.append(ref)
        sys.stdout.write('weakref: %s\n' % (ref(),))
        
    def Notify(self):
        for o in self.observers:
            o = o()
            
            
            try:
                import gc
            except ImportError:
                o = None #some jython does not have gc, so, there's no sense testing this in it
            else:
                try:
                    gc.get_referrers(o)
                except:
                    o = None #jython and ironpython do not have get_referrers
            
            if o is not None:
                sys.stdout.write('still observing %s\n' % (o,))
                sys.stdout.write('number of referrers: %s\n' % len(gc.get_referrers(o)))
                frame = gc.get_referrers(o)[0]
                frame_referrers = gc.get_referrers(frame)
                sys.stdout.write('frame referrer %s\n' % (frame_referrers,))
                referrers1 = gc.get_referrers(frame_referrers[1])
                sys.stdout.write('%s\n' % (referrers1,))
                sys.stderr.write('TEST FAILED: The observer should have died, even when running in debug\n')
            else:
                sys.stdout.write('TEST SUCEEDED: observer died\n')
                
            sys.stdout.flush()
            sys.stderr.flush()
                
class Observer(object):
    pass

    
def main():
    observable = set_up()
    observable.Notify()
    
    
if __name__ == '__main__':
    main()



from typing import Dict, Tuple, Sequence, Iterator, List
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import ICompletionContext
from robocorp_ls_core.lsp import (
    CompletionItem,
    CompletionItemKind,
    InsertTextFormat,
    Position,
    Range,
    TextEdit,
    CompletionItemTypedDict,
)

log = get_logger(__name__)


def _iter_normalized_variables_and_values(
    completion_context: ICompletionContext,
) -> Iterator[Tuple[str, Tuple[str, ...]]]:
    from robot.api import Token
    from robotframework_ls.impl.variable_resolve import robot_search_variable
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    for node_info in completion_context.get_all_variables():
        node = node_info.node
        token = node.get_token(Token.VARIABLE)
        if token is None:
            continue
        var_name = token.value
        robot_match = robot_search_variable(var_name)
        if robot_match and robot_match.base:
            # i.e.: Variable.value provides the values of the assign
            var_value: Tuple[str, ...] = node.value
            yield (normalize_robot_name(robot_match.base), var_value)


def _as_dictionary(
    dict_tokens: Sequence[str], normalize=False, filter_token: str = ""
) -> Dict[str, str]:
    """
    Parse ["key1=val1", "key2=val2",...] as a dictionary
    """
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    dictionary = {}
    for token in dict_tokens:
        key, val = token.split("=")
        if normalize:
            key = normalize_robot_name(key)
        if filter_token and filter_token not in normalize_robot_name(key):
            continue
        dictionary.update({key: val})
    return dictionary


def _completion_items(
    dictionary: Dict[str, str], editor_range: Range
) -> List[CompletionItemTypedDict]:
    return [
        CompletionItem(
            key,
            kind=CompletionItemKind.Variable,
            text_edit=TextEdit(editor_range, key),
            insertText=key,
            documentation=value,
            insertTextFormat=InsertTextFormat.Snippet,
        ).to_dict()
        for key, value in dictionary.items()
    ]


def _iter_all_normalized_variables_and_values(
    completion_context: ICompletionContext,
) -> Iterator[Tuple[str, Tuple[str, ...]]]:

    yield from _iter_normalized_variables_and_values(completion_context)

    dependency_graph = completion_context.collect_dependency_graph()
    for resource_doc in completion_context.iter_dependency_and_init_resource_docs(
        dependency_graph
    ):
        new_ctx = completion_context.create_copy(resource_doc)
        yield from _iter_normalized_variables_and_values(new_ctx)


def complete(completion_context: ICompletionContext) -> List[CompletionItemTypedDict]:
    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches
    from robotframework_ls.impl.ast_utils import iter_robot_match_as_tokens
    from robotframework_ls.impl.text_utilities import normalize_robot_name
    from robotframework_ls.impl.variable_resolve import robot_search_variable

    token_info = completion_context.get_current_token()
    if token_info is None:
        return []
    token = token_info.token
    value = token.value

    col = completion_context.sel.col

    last_opening_bracket_column = -1

    items_seen = []

    prev_rtoken = None
    for robot_match, relative_index in iter_robot_variable_matches(value):
        robot_match_start = token.col_offset + relative_index + robot_match.start
        robot_match_end = token.col_offset + relative_index + robot_match.end

        if robot_match.base and robot_match_start < col < robot_match_end:
            # Now, let's see in which item/offset we're in.
            for rtoken in iter_robot_match_as_tokens(
                robot_match, relative_index=robot_match_start, lineno=token.lineno
            ):
                if rtoken.type == "[":
                    last_opening_bracket_column = rtoken.col_offset

                if rtoken.col_offset >= col:
                    if (
                        rtoken.type == "item"
                        and rtoken.col_offset == rtoken.end_col_offset
                    ):
                        # dealing with empty item at cursor
                        prev_rtoken = rtoken
                        items_seen.append(rtoken)

                    # The prev_rtoken is the one we're interested in
                    break

                if rtoken.type == "item":
                    items_seen.append(rtoken)

                prev_rtoken = rtoken

            break
    else:
        return []

    if prev_rtoken is None:
        return []

    if prev_rtoken.type not in ("[", "item"):
        return []

    if last_opening_bracket_column == -1:
        return []

    search_items_normalized = [normalize_robot_name(robot_match.base)]
    if len(items_seen) > 1:
        for item in items_seen[:-1]:
            # The last one is the one we're currently completing
            search_items_normalized.append(normalize_robot_name(item.value))

    selection = completion_context.sel

    if prev_rtoken.type == "[":
        start_offset = end_offset = prev_rtoken.col_offset
        filter_token = ""
    else:
        start_offset = prev_rtoken.col_offset
        end_offset = prev_rtoken.end_col_offset
        filter_token = normalize_robot_name(prev_rtoken.value)

    normalized_variables_and_values = dict(
        _iter_all_normalized_variables_and_values(completion_context)
    )

    last_dict = None
    count = 0
    while search_items_normalized:
        count += 1
        if count > 10:
            log.info(
                "Breaking up possible recursion on dictionary completions. Stack: %s",
                search_items_normalized,
            )
            return []
        search_name_normalized = search_items_normalized.pop(0)

        variable_values = normalized_variables_and_values.get(search_name_normalized)
        if not variable_values:
            return []
        if not search_items_normalized:
            dictionary = _as_dictionary(variable_values, filter_token=filter_token)
            editor_range = Range(
                start=Position(selection.line, start_offset),
                end=Position(selection.line, end_offset),
            )
            return _completion_items(dictionary, editor_range)
        else:
            last_dict = _as_dictionary(variable_values, normalize=True)

            next_search = last_dict.get(search_items_normalized.pop(0))
            if not next_search:
                return []

            if not next_search.startswith("&{"):
                return []

            new_match = robot_search_variable(next_search)
            if not new_match or not new_match.base:
                return []

            for it in reversed(new_match.items):
                search_items_normalized.insert(0, normalize_robot_name(it))
            search_items_normalized.insert(0, normalize_robot_name(new_match.base))

    return []




class Car(object):
    """A car class"""
    def __init__(self, model, make, color):
        self.model = model
        self.make = make
        self.color = color
        self.price = None

    def get_price(self):
        return self.price

    def set_price(self, value):
        self.price = value

availableCars = []
def main():
    global availableCars

    #Create a new car obj
    carObj = Car("Maruti SX4", "2011", "Black")
    carObj.set_price(950000)  # Set price
    # Add this to available cars
    availableCars.append(carObj)

    print('TEST SUCEEDED')

if __name__ == '__main__':
    main()



import time


def check_output_command(messages_sent):
    for _i in range(50):
        for msg in messages_sent:
            if msg.get('type') == 'event' and msg.get('event') == 'output' and 'Something' in msg['body']['output']:
                return True
        time.sleep(.1)
    return False


def check_continue_request(messages_received):
    for _i in range(50):
        for msg in messages_received:
            if msg.get('type') == 'request' and msg.get('command') == 'continue':
                return True
        time.sleep(.1)
    return False


def main():
    '''
    In this test we'll check that pydevd sends an output event and receives a continue request in
    the dap messages listener.
    '''
    messages_sent = []
    messages_received = []

    import pydevd

    class Listener(pydevd.IDAPMessagesListener):

        def before_send(self, message_as_dict):
            messages_sent.append(message_as_dict)

        def after_receive(self, message_as_dict):
            messages_received.append(message_as_dict)

    pydevd.add_dap_messages_listener(Listener())

    print('Something')  # Break here

    # Note that the message is only received in a thread, so, we have to
    # wait for it to be received.
    if not check_output_command(messages_sent):
        print('Output not received!')
        return False

    if not check_continue_request(messages_received):
        print('Continue not received!')
        return False

    print('TEST SUCEEDED!')
    return True


if __name__ == '__main__':
    assert main()



import os.path

from robocorp_ls_core.cache import instance_cache
from robocorp_ls_core.protocols import check_implements
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    IKeywordFound,
    ICompletionContext,
    IKeywordCollector,
    IKeywordArg,
    LibraryDependencyInfo,
    AbstractKeywordCollector,
)
from typing import Sequence, List, Dict, Optional, Iterator
from robotframework_ls.impl.text_utilities import build_keyword_docs_with_signature
from robocorp_ls_core.lsp import MarkupContentTypedDict, MarkupKind


log = get_logger(__name__)


class _KeywordFoundFromAst(object):

    __slots__ = [
        "_module_ast",
        "_keyword_node",
        "_keyword_name",
        "_keyword_args",
        "completion_context",
        "completion_item_kind",
        "_name_token",
        "__instance_cache__",
    ]

    def __init__(
        self,
        module_ast,
        keyword_node,
        keyword_name,
        keyword_args: Sequence[IKeywordArg],
        completion_context,
        completion_item_kind,
    ):
        from robot.api import Token

        self._module_ast = module_ast
        self._keyword_node = keyword_node

        self._name_token = keyword_node.header.get_token(Token.KEYWORD_NAME)

        self._keyword_name = keyword_name
        self._keyword_args = keyword_args
        self.completion_context = completion_context
        self.completion_item_kind = completion_item_kind

    @property
    def keyword_name(self):
        return self._keyword_name

    @property
    def keyword_args(self) -> Sequence[IKeywordArg]:
        return self._keyword_args

    @property
    def library_alias(self):
        return None

    @property
    def library_name(self):
        return None

    @property
    def resource_name(self):
        from robocorp_ls_core import uris

        uri = self.completion_context.doc.uri
        return os.path.splitext(os.path.basename(uris.to_fs_path(uri)))[0]

    def compute_docs_without_signature(self) -> MarkupContentTypedDict:
        return {
            "kind": MarkupKind.Markdown,
            "value": self._docs_without_signature(),
        }

    def compute_docs_with_signature(self) -> MarkupContentTypedDict:
        docs = build_keyword_docs_with_signature(
            self.keyword_name,
            tuple(x.original_arg for x in self.keyword_args),
            self._docs_without_signature(),
            "markdown",
        )
        return {
            "kind": MarkupKind.Markdown,
            "value": docs,
        }

    @instance_cache
    def _docs_without_signature(self) -> str:
        from robotframework_ls.impl import ast_utils

        return ast_utils.get_documentation_as_markdown(self._keyword_node)

    @instance_cache
    def is_deprecated(self) -> bool:
        from robotframework_ls.impl import ast_utils

        return ast_utils.is_deprecated(self._keyword_node)

    @property  # type: ignore
    @instance_cache
    def source(self) -> str:
        from robocorp_ls_core import uris

        return uris.to_fs_path(self.completion_context.doc.uri)

    @property
    def lineno(self):
        return self._name_token.lineno - 1

    @property
    def end_lineno(self):
        return self._name_token.lineno - 1

    @property
    def col_offset(self):
        return self._name_token.col_offset

    @property
    def end_col_offset(self):
        return self._name_token.end_col_offset

    @property
    def scope_lineno(self) -> Optional[int]:
        return self._keyword_node.lineno - 1

    @property
    def scope_end_lineno(self) -> Optional[int]:
        return self._keyword_node.end_lineno - 1

    @property
    def scope_col_offset(self) -> Optional[int]:
        return self._keyword_node.col_offset

    @property
    def scope_end_col_offset(self) -> Optional[int]:
        return self._keyword_node.end_col_offset

    def __typecheckself__(self) -> None:
        _: IKeywordFound = check_implements(self)


class _KeywordFoundFromLibrary(object):

    __slots__ = [
        "_library_doc",
        "_library_alias",
        "_keyword_doc",
        "_keyword_name",
        "_keyword_args",
        "completion_context",
        "completion_item_kind",
        "__instance_cache__",
    ]

    def __init__(
        self,
        library_doc,
        keyword_doc,
        keyword_name,
        keyword_args: Sequence[IKeywordArg],
        completion_context,
        completion_item_kind,
        library_alias=None,
    ):

        self._library_doc = library_doc
        self._keyword_doc = keyword_doc
        self._keyword_name = keyword_name
        self._keyword_args = keyword_args

        self.completion_context = completion_context
        self.completion_item_kind = completion_item_kind
        self._library_alias = library_alias

    @property
    def keyword_name(self):
        return self._keyword_name

    @property
    def keyword_args(self) -> Sequence[IKeywordArg]:
        return self._keyword_args

    @property
    def library_alias(self):
        return self._library_alias

    @property
    def library_name(self):
        return self._library_doc.name

    @property
    def resource_name(self):
        return None

    @instance_cache
    def is_deprecated(self):
        from robotframework_ls.impl import text_utilities

        return text_utilities.has_deprecated_text(self._keyword_doc.doc)

    @property  # type: ignore
    @instance_cache
    def source(self):
        source = self._keyword_doc.source or self._library_doc.source
        return source

    @property  # type: ignore
    @instance_cache
    def lineno(self):
        return self._keyword_doc.lineno - 1  # i.e.: make 0-based.

    @property
    def end_lineno(self):
        return self.lineno

    @property
    def col_offset(self):
        return 0

    @property
    def end_col_offset(self):
        return 0

    @property
    def scope_lineno(self) -> Optional[int]:
        return self.lineno

    @property
    def scope_end_lineno(self) -> Optional[int]:
        return self.end_lineno

    @property
    def scope_col_offset(self) -> Optional[int]:
        return self.col_offset

    @property
    def scope_end_col_offset(self) -> Optional[int]:
        return self.end_col_offset

    def compute_docs_without_signature(self) -> MarkupContentTypedDict:
        from robotframework_ls.impl.robot_specbuilder import docs_and_format

        docs, docs_format = docs_and_format(self._keyword_doc)

        return {"kind": docs_format, "value": docs}

    def compute_docs_with_signature(self) -> MarkupContentTypedDict:
        from robotframework_ls.impl.robot_specbuilder import docs_and_format

        docs, docs_format = docs_and_format(self._keyword_doc)
        docs = build_keyword_docs_with_signature(
            self.keyword_name,
            tuple(x.original_arg for x in self.keyword_args),
            docs,
            docs_format,
        )

        return {
            "kind": docs_format,
            "value": docs,
        }

    def __typecheckself__(self) -> None:
        _: IKeywordFound = check_implements(self)


def _collect_completions_from_ast(
    ast, completion_context: ICompletionContext, collector
):
    from robotframework_ls.impl import ast_utils
    from robocorp_ls_core.lsp import CompletionItemKind

    found = {}

    for keyword in ast_utils.iter_keywords(ast):
        completion_context.check_cancelled()
        keyword_name = keyword.node.name
        if collector.accepts(keyword_name):
            keyword_args = list(ast_utils.iter_keyword_arguments_as_kwarg(keyword.node))

            found[keyword_name] = _KeywordFoundFromAst(
                ast,
                keyword.node,
                keyword_name,
                keyword_args,
                completion_context,
                CompletionItemKind.Function,
            )

    # We notify afterwards because if multiple definitions of the same
    # keyword are found, we just want to report the last one (as is the
    # case for the interactive console).
    for keyword_found in found.values():
        collector.on_keyword(keyword_found)


def _collect_current_doc_keywords(
    completion_context: ICompletionContext, collector: IKeywordCollector
):
    """
    :param CompletionContext completion_context:
    """
    # Get keywords defined in the file itself

    ast = completion_context.get_ast()
    _collect_completions_from_ast(ast, completion_context, collector)


def _collect_libraries_keywords(
    completion_context: ICompletionContext,
    library_infos: Iterator[LibraryDependencyInfo],
    collector: IKeywordCollector,
):
    from robotframework_ls.impl.libspec_manager import LibspecManager
    from robotframework_ls.impl.protocols import ILibraryDocOrError

    # Get keywords from libraries
    from robocorp_ls_core.lsp import CompletionItemKind

    libspec_manager: LibspecManager = completion_context.workspace.libspec_manager

    for library_info in library_infos:
        completion_context.check_cancelled()

        library_doc_or_error: ILibraryDocOrError = (
            libspec_manager.get_library_doc_or_error(
                library_info.name,
                create=True,
                completion_context=completion_context,
                builtin=library_info.builtin,
                args=library_info.args,
            )
        )
        library_doc = library_doc_or_error.library_doc
        if library_doc is not None:
            #: :type keyword: KeywordDoc
            for keyword in library_doc.keywords:
                keyword_name = keyword.name
                if collector.accepts(keyword_name):

                    keyword_args: Sequence[IKeywordArg] = ()
                    if keyword.args:
                        keyword_args = keyword.args

                    collector.on_keyword(
                        _KeywordFoundFromLibrary(
                            library_doc,
                            keyword,
                            keyword_name,
                            keyword_args,
                            completion_context,
                            CompletionItemKind.Method,
                            library_alias=library_info.alias,
                        )
                    )

            collector.on_resolved_library(
                completion_context, library_info.node, library_doc
            )
        else:
            from robot.api import Token

            error_msg = library_doc_or_error.error
            node = library_info.node
            if node:
                if error_msg:
                    error_msg = f"\nError generating libspec:\n{error_msg}"

                node_name_tok = node.get_token(Token.NAME)

                if node_name_tok is not None:
                    (
                        _value,
                        token_errors,
                    ) = completion_context.token_value_and_unresolved_resolving_variables(
                        node_name_tok
                    )
                    if token_errors:
                        for token_error in token_errors:
                            collector.on_unresolved_library(
                                completion_context,
                                node.name,
                                token_error.lineno,
                                token_error.lineno,
                                token_error.col_offset,
                                token_error.end_col_offset,
                                f"\nUnable to statically resolve variable: {token_error.value}.\nPlease set the `{token_error.value[2:-1]}` value in `robot.variables`.",
                            )
                    else:
                        collector.on_unresolved_library(
                            completion_context,
                            node.name,
                            node_name_tok.lineno,
                            node_name_tok.lineno,
                            node_name_tok.col_offset,
                            node_name_tok.end_col_offset,
                            error_msg,
                        )
                else:
                    collector.on_unresolved_library(
                        completion_context,
                        library_info.name,
                        node.lineno,
                        node.end_lineno,
                        node.col_offset,
                        node.end_col_offset,
                        error_msg,
                    )


def _collect_from_context(
    completion_context: ICompletionContext, collector: IKeywordCollector
):
    dependency_graph = completion_context.collect_dependency_graph()

    root_doc = dependency_graph.get_root_doc()
    assert root_doc.uri == completion_context.doc.uri
    _collect_current_doc_keywords(completion_context, collector)

    completion_context.check_cancelled()
    _collect_libraries_keywords(
        completion_context,
        dependency_graph.iter_libraries(completion_context.doc.uri),
        collector,
    )

    for node, resource_doc in dependency_graph.iter_all_resource_imports_with_docs():
        completion_context.check_cancelled()

        if resource_doc is None:
            # Note that 'None' documents will only be given for the
            # initial context (so, it's ok to use `completion_context`
            # in this case).
            from robot.api import Token

            node_name_tok = node.get_token(Token.NAME)
            if node_name_tok is not None:

                (
                    _value,
                    token_errors,
                ) = completion_context.token_value_and_unresolved_resolving_variables(
                    node_name_tok
                )

                if token_errors:
                    for token_error in token_errors:
                        collector.on_unresolved_resource(
                            completion_context,
                            node.name,
                            token_error.lineno,
                            token_error.lineno,
                            token_error.col_offset,
                            token_error.end_col_offset,
                            f"\nUnable to statically resolve variable: {token_error.value}.\nPlease set the `{token_error.value[2:-1]}` value in `robot.variables`.",
                        )

                else:
                    collector.on_unresolved_resource(
                        completion_context,
                        node.name,
                        node_name_tok.lineno,
                        node_name_tok.lineno,
                        node_name_tok.col_offset,
                        node_name_tok.end_col_offset,
                        None,
                    )
            else:
                collector.on_unresolved_resource(
                    completion_context,
                    node.name,
                    node.lineno,
                    node.end_lineno,
                    node.col_offset,
                    node.end_col_offset,
                    None,
                )
            continue
        completion_context.check_cancelled()
        new_ctx = completion_context.create_copy(resource_doc)

        _collect_current_doc_keywords(new_ctx, collector)
        _collect_libraries_keywords(
            new_ctx,
            dependency_graph.iter_libraries(resource_doc.uri),
            collector,
        )


class _CollectKeywordNameToKeywordFound(AbstractKeywordCollector):
    def __init__(self) -> None:
        self.keyword_name_to_keyword_found: Dict[str, List[IKeywordFound]] = {}

    def accepts(self, keyword_name: str) -> bool:
        return True

    def on_keyword(self, keyword_found: IKeywordFound):
        lst = self.keyword_name_to_keyword_found.get(keyword_found.keyword_name)
        if lst is None:
            self.keyword_name_to_keyword_found[keyword_found.keyword_name] = lst = []
        lst.append(keyword_found)

    def __typecheckself__(self) -> None:
        _: IKeywordCollector = check_implements(self)


def collect_keyword_name_to_keyword_found(
    completion_context: ICompletionContext,
) -> Dict[str, List[IKeywordFound]]:
    collector = _CollectKeywordNameToKeywordFound()
    _collect_from_context(completion_context, collector)
    return collector.keyword_name_to_keyword_found


def collect_keywords(
    completion_context: ICompletionContext, collector: IKeywordCollector
):
    """
    Collects all the keywords that are available to the given completion_context.
    """
    _collect_from_context(completion_context, collector)



def method():
    a = 1
    print('call %s' % (a,))

    def method2():
        print('call %s' % (a,))

    while a < 10:
        a += 1
        print('call %s' % (a,))

    try:
        if a < 0:
            print('call %s' % (a,))
            raise ValueError
        else:
            method2()
    except ValueError:
        pass

    print('call %s' % (a,))


if __name__ == '__main__':
    method()
    print('TEST SUCEEDED!')


import sys
import os


def main():
    env = os.environ.copy()
    pythonpath = env.get('PYTHONPATH', '')
    env['PYTHONPATH'] = os.path.dirname(__file__) + os.pathsep + \
        os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

    from _pydevd_bundle.pydevd_constants import HTTP_JSON_PROTOCOL
    from _pydevd_bundle.pydevd_defaults import PydevdCustomization
    PydevdCustomization.DEFAULT_PROTOCOL = HTTP_JSON_PROTOCOL

    import pydevd
    from _pydev_bundle import pydev_log
    pydev_log.debug('Argv received: %s', sys.argv)
    port = int(sys.argv[1])
    print('before pydevd.settrace')
    pydevd.settrace(port=port, patch_multiprocessing=True, suspend=True)
    print('after pydevd.settrace')

    import subprocess
    if '--use-c-switch' in sys.argv:
        child_process = subprocess.Popen(
            [sys.executable, '-u', '-c', 'import _debugger_case_pydevd_customization;_debugger_case_pydevd_customization.call()'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
    elif '--posix-spawn' in sys.argv:
        env = os.environ.copy()
        args = ['-u', '_debugger_case_pydevd_customization.py', '--simple-call']
        pid = os.posix_spawn(sys.executable, args, env)
        os.waitpid(pid, 0)
        child_process = None  # We don't really have a subprocess.Popen instance in this case.
    else:
        child_process = subprocess.Popen(
            [sys.executable, '-u', '_debugger_case_pydevd_customization.py', '--simple-call'],
            cwd=os.path.dirname(__file__),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )

    if child_process:
        stdout, stderr = child_process.communicate()
        assert b'called' in stdout, 'Did not find b"called" in stdout:\n>>%s<<\nstderr:\n>>%s<<\n' % (stdout, stderr)
    print('TEST SUCEEDED!')  # break 2 here


def call():
    import pydevd
    from _pydevd_bundle.pydevd_api import PyDevdAPI
    assert pydevd.get_global_debugger().get_arg_ppid() == PyDevdAPI().get_ppid()
    print("called")  # break 1 here


if __name__ == '__main__':
    if '--simple-call' in sys.argv:
        call()
    else:
        main()



def Call():
    variables_for_test_1 = 100
    variables_for_test_2 = [1, 10, 100]
    variables_for_test_3 = {10: 10, 100: 100, 1000: 1000}
    variables_for_test_4 = {(1, 10, 100): (10000, 100000, 100000)}

    all_vars_set = True  # Break here


if __name__ == '__main__':
    Call()
    print('TEST SUCEEDED!')


"""Jinja is a template engine written in pure Python. It provides a
non-XML syntax that supports inline expressions and an optional
sandboxed environment.
"""
from .bccache import BytecodeCache as BytecodeCache
from .bccache import FileSystemBytecodeCache as FileSystemBytecodeCache
from .bccache import MemcachedBytecodeCache as MemcachedBytecodeCache
from .environment import Environment as Environment
from .environment import Template as Template
from .exceptions import TemplateAssertionError as TemplateAssertionError
from .exceptions import TemplateError as TemplateError
from .exceptions import TemplateNotFound as TemplateNotFound
from .exceptions import TemplateRuntimeError as TemplateRuntimeError
from .exceptions import TemplatesNotFound as TemplatesNotFound
from .exceptions import TemplateSyntaxError as TemplateSyntaxError
from .exceptions import UndefinedError as UndefinedError
from .loaders import BaseLoader as BaseLoader
from .loaders import ChoiceLoader as ChoiceLoader
from .loaders import DictLoader as DictLoader
from .loaders import FileSystemLoader as FileSystemLoader
from .loaders import FunctionLoader as FunctionLoader
from .loaders import ModuleLoader as ModuleLoader
from .loaders import PackageLoader as PackageLoader
from .loaders import PrefixLoader as PrefixLoader
from .runtime import ChainableUndefined as ChainableUndefined
from .runtime import DebugUndefined as DebugUndefined
from .runtime import make_logging_undefined as make_logging_undefined
from .runtime import StrictUndefined as StrictUndefined
from .runtime import Undefined as Undefined
from .utils import clear_caches as clear_caches
from .utils import is_undefined as is_undefined
from .utils import pass_context as pass_context
from .utils import pass_environment as pass_environment
from .utils import pass_eval_context as pass_eval_context
from .utils import select_autoescape as select_autoescape

__version__ = "3.1.1"



"""A sandbox layer that ensures unsafe operations cannot be performed.
Useful when the template itself comes from an untrusted source.
"""
import operator
import types
import typing as t
from _string import formatter_field_name_split  # type: ignore
from collections import abc
from collections import deque
from string import Formatter

from markupsafe import EscapeFormatter
from markupsafe import Markup

from .environment import Environment
from .exceptions import SecurityError
from .runtime import Context
from .runtime import Undefined

F = t.TypeVar("F", bound=t.Callable[..., t.Any])

#: maximum number of items a range may produce
MAX_RANGE = 100000

#: Unsafe function attributes.
UNSAFE_FUNCTION_ATTRIBUTES: t.Set[str] = set()

#: Unsafe method attributes. Function attributes are unsafe for methods too.
UNSAFE_METHOD_ATTRIBUTES: t.Set[str] = set()

#: unsafe generator attributes.
UNSAFE_GENERATOR_ATTRIBUTES = {"gi_frame", "gi_code"}

#: unsafe attributes on coroutines
UNSAFE_COROUTINE_ATTRIBUTES = {"cr_frame", "cr_code"}

#: unsafe attributes on async generators
UNSAFE_ASYNC_GENERATOR_ATTRIBUTES = {"ag_code", "ag_frame"}

_mutable_spec: t.Tuple[t.Tuple[t.Type, t.FrozenSet[str]], ...] = (
    (
        abc.MutableSet,
        frozenset(
            [
                "add",
                "clear",
                "difference_update",
                "discard",
                "pop",
                "remove",
                "symmetric_difference_update",
                "update",
            ]
        ),
    ),
    (
        abc.MutableMapping,
        frozenset(["clear", "pop", "popitem", "setdefault", "update"]),
    ),
    (
        abc.MutableSequence,
        frozenset(["append", "reverse", "insert", "sort", "extend", "remove"]),
    ),
    (
        deque,
        frozenset(
            [
                "append",
                "appendleft",
                "clear",
                "extend",
                "extendleft",
                "pop",
                "popleft",
                "remove",
                "rotate",
            ]
        ),
    ),
)


def inspect_format_method(callable: t.Callable) -> t.Optional[str]:
    if not isinstance(
        callable, (types.MethodType, types.BuiltinMethodType)
    ) or callable.__name__ not in ("format", "format_map"):
        return None

    obj = callable.__self__

    if isinstance(obj, str):
        return obj

    return None


def safe_range(*args: int) -> range:
    """A range that can't generate ranges with a length of more than
    MAX_RANGE items.
    """
    rng = range(*args)

    if len(rng) > MAX_RANGE:
        raise OverflowError(
            "Range too big. The sandbox blocks ranges larger than"
            f" MAX_RANGE ({MAX_RANGE})."
        )

    return rng


def unsafe(f: F) -> F:
    """Marks a function or method as unsafe.

    .. code-block: python

        @unsafe
        def delete(self):
            pass
    """
    f.unsafe_callable = True  # type: ignore
    return f


def is_internal_attribute(obj: t.Any, attr: str) -> bool:
    """Test if the attribute given is an internal python attribute.  For
    example this function returns `True` for the `func_code` attribute of
    python objects.  This is useful if the environment method
    :meth:`~SandboxedEnvironment.is_safe_attribute` is overridden.

    >>> from jinja2.sandbox import is_internal_attribute
    >>> is_internal_attribute(str, "mro")
    True
    >>> is_internal_attribute(str, "upper")
    False
    """
    if isinstance(obj, types.FunctionType):
        if attr in UNSAFE_FUNCTION_ATTRIBUTES:
            return True
    elif isinstance(obj, types.MethodType):
        if attr in UNSAFE_FUNCTION_ATTRIBUTES or attr in UNSAFE_METHOD_ATTRIBUTES:
            return True
    elif isinstance(obj, type):
        if attr == "mro":
            return True
    elif isinstance(obj, (types.CodeType, types.TracebackType, types.FrameType)):
        return True
    elif isinstance(obj, types.GeneratorType):
        if attr in UNSAFE_GENERATOR_ATTRIBUTES:
            return True
    elif hasattr(types, "CoroutineType") and isinstance(obj, types.CoroutineType):
        if attr in UNSAFE_COROUTINE_ATTRIBUTES:
            return True
    elif hasattr(types, "AsyncGeneratorType") and isinstance(
        obj, types.AsyncGeneratorType
    ):
        if attr in UNSAFE_ASYNC_GENERATOR_ATTRIBUTES:
            return True
    return attr.startswith("__")


def modifies_known_mutable(obj: t.Any, attr: str) -> bool:
    """This function checks if an attribute on a builtin mutable object
    (list, dict, set or deque) or the corresponding ABCs would modify it
    if called.

    >>> modifies_known_mutable({}, "clear")
    True
    >>> modifies_known_mutable({}, "keys")
    False
    >>> modifies_known_mutable([], "append")
    True
    >>> modifies_known_mutable([], "index")
    False

    If called with an unsupported object, ``False`` is returned.

    >>> modifies_known_mutable("foo", "upper")
    False
    """
    for typespec, unsafe in _mutable_spec:
        if isinstance(obj, typespec):
            return attr in unsafe
    return False


class SandboxedEnvironment(Environment):
    """The sandboxed environment.  It works like the regular environment but
    tells the compiler to generate sandboxed code.  Additionally subclasses of
    this environment may override the methods that tell the runtime what
    attributes or functions are safe to access.

    If the template tries to access insecure code a :exc:`SecurityError` is
    raised.  However also other exceptions may occur during the rendering so
    the caller has to ensure that all exceptions are caught.
    """

    sandboxed = True

    #: default callback table for the binary operators.  A copy of this is
    #: available on each instance of a sandboxed environment as
    #: :attr:`binop_table`
    default_binop_table: t.Dict[str, t.Callable[[t.Any, t.Any], t.Any]] = {
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,
        "/": operator.truediv,
        "//": operator.floordiv,
        "**": operator.pow,
        "%": operator.mod,
    }

    #: default callback table for the unary operators.  A copy of this is
    #: available on each instance of a sandboxed environment as
    #: :attr:`unop_table`
    default_unop_table: t.Dict[str, t.Callable[[t.Any], t.Any]] = {
        "+": operator.pos,
        "-": operator.neg,
    }

    #: a set of binary operators that should be intercepted.  Each operator
    #: that is added to this set (empty by default) is delegated to the
    #: :meth:`call_binop` method that will perform the operator.  The default
    #: operator callback is specified by :attr:`binop_table`.
    #:
    #: The following binary operators are interceptable:
    #: ``//``, ``%``, ``+``, ``*``, ``-``, ``/``, and ``**``
    #:
    #: The default operation form the operator table corresponds to the
    #: builtin function.  Intercepted calls are always slower than the native
    #: operator call, so make sure only to intercept the ones you are
    #: interested in.
    #:
    #: .. versionadded:: 2.6
    intercepted_binops: t.FrozenSet[str] = frozenset()

    #: a set of unary operators that should be intercepted.  Each operator
    #: that is added to this set (empty by default) is delegated to the
    #: :meth:`call_unop` method that will perform the operator.  The default
    #: operator callback is specified by :attr:`unop_table`.
    #:
    #: The following unary operators are interceptable: ``+``, ``-``
    #:
    #: The default operation form the operator table corresponds to the
    #: builtin function.  Intercepted calls are always slower than the native
    #: operator call, so make sure only to intercept the ones you are
    #: interested in.
    #:
    #: .. versionadded:: 2.6
    intercepted_unops: t.FrozenSet[str] = frozenset()

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.globals["range"] = safe_range
        self.binop_table = self.default_binop_table.copy()
        self.unop_table = self.default_unop_table.copy()

    def is_safe_attribute(self, obj: t.Any, attr: str, value: t.Any) -> bool:
        """The sandboxed environment will call this method to check if the
        attribute of an object is safe to access.  Per default all attributes
        starting with an underscore are considered private as well as the
        special attributes of internal python objects as returned by the
        :func:`is_internal_attribute` function.
        """
        return not (attr.startswith("_") or is_internal_attribute(obj, attr))

    def is_safe_callable(self, obj: t.Any) -> bool:
        """Check if an object is safely callable. By default callables
        are considered safe unless decorated with :func:`unsafe`.

        This also recognizes the Django convention of setting
        ``func.alters_data = True``.
        """
        return not (
            getattr(obj, "unsafe_callable", False) or getattr(obj, "alters_data", False)
        )

    def call_binop(
        self, context: Context, operator: str, left: t.Any, right: t.Any
    ) -> t.Any:
        """For intercepted binary operator calls (:meth:`intercepted_binops`)
        this function is executed instead of the builtin operator.  This can
        be used to fine tune the behavior of certain operators.

        .. versionadded:: 2.6
        """
        return self.binop_table[operator](left, right)

    def call_unop(self, context: Context, operator: str, arg: t.Any) -> t.Any:
        """For intercepted unary operator calls (:meth:`intercepted_unops`)
        this function is executed instead of the builtin operator.  This can
        be used to fine tune the behavior of certain operators.

        .. versionadded:: 2.6
        """
        return self.unop_table[operator](arg)

    def getitem(
        self, obj: t.Any, argument: t.Union[str, t.Any]
    ) -> t.Union[t.Any, Undefined]:
        """Subscribe an object from sandboxed code."""
        try:
            return obj[argument]
        except (TypeError, LookupError):
            if isinstance(argument, str):
                try:
                    attr = str(argument)
                except Exception:
                    pass
                else:
                    try:
                        value = getattr(obj, attr)
                    except AttributeError:
                        pass
                    else:
                        if self.is_safe_attribute(obj, argument, value):
                            return value
                        return self.unsafe_undefined(obj, argument)
        return self.undefined(obj=obj, name=argument)

    def getattr(self, obj: t.Any, attribute: str) -> t.Union[t.Any, Undefined]:
        """Subscribe an object from sandboxed code and prefer the
        attribute.  The attribute passed *must* be a bytestring.
        """
        try:
            value = getattr(obj, attribute)
        except AttributeError:
            try:
                return obj[attribute]
            except (TypeError, LookupError):
                pass
        else:
            if self.is_safe_attribute(obj, attribute, value):
                return value
            return self.unsafe_undefined(obj, attribute)
        return self.undefined(obj=obj, name=attribute)

    def unsafe_undefined(self, obj: t.Any, attribute: str) -> Undefined:
        """Return an undefined object for unsafe attributes."""
        return self.undefined(
            f"access to attribute {attribute!r} of"
            f" {type(obj).__name__!r} object is unsafe.",
            name=attribute,
            obj=obj,
            exc=SecurityError,
        )

    def format_string(
        self,
        s: str,
        args: t.Tuple[t.Any, ...],
        kwargs: t.Dict[str, t.Any],
        format_func: t.Optional[t.Callable] = None,
    ) -> str:
        """If a format call is detected, then this is routed through this
        method so that our safety sandbox can be used for it.
        """
        formatter: SandboxedFormatter
        if isinstance(s, Markup):
            formatter = SandboxedEscapeFormatter(self, escape=s.escape)
        else:
            formatter = SandboxedFormatter(self)

        if format_func is not None and format_func.__name__ == "format_map":
            if len(args) != 1 or kwargs:
                raise TypeError(
                    "format_map() takes exactly one argument"
                    f" {len(args) + (kwargs is not None)} given"
                )

            kwargs = args[0]
            args = ()

        rv = formatter.vformat(s, args, kwargs)
        return type(s)(rv)

    def call(
        __self,  # noqa: B902
        __context: Context,
        __obj: t.Any,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> t.Any:
        """Call an object from sandboxed code."""
        fmt = inspect_format_method(__obj)
        if fmt is not None:
            return __self.format_string(fmt, args, kwargs, __obj)

        # the double prefixes are to avoid double keyword argument
        # errors when proxying the call.
        if not __self.is_safe_callable(__obj):
            raise SecurityError(f"{__obj!r} is not safely callable")
        return __context.call(__obj, *args, **kwargs)


class ImmutableSandboxedEnvironment(SandboxedEnvironment):
    """Works exactly like the regular `SandboxedEnvironment` but does not
    permit modifications on the builtin mutable objects `list`, `set`, and
    `dict` by using the :func:`modifies_known_mutable` function.
    """

    def is_safe_attribute(self, obj: t.Any, attr: str, value: t.Any) -> bool:
        if not super().is_safe_attribute(obj, attr, value):
            return False

        return not modifies_known_mutable(obj, attr)


class SandboxedFormatter(Formatter):
    def __init__(self, env: Environment, **kwargs: t.Any) -> None:
        self._env = env
        super().__init__(**kwargs)

    def get_field(
        self, field_name: str, args: t.Sequence[t.Any], kwargs: t.Mapping[str, t.Any]
    ) -> t.Tuple[t.Any, str]:
        first, rest = formatter_field_name_split(field_name)
        obj = self.get_value(first, args, kwargs)
        for is_attr, i in rest:
            if is_attr:
                obj = self._env.getattr(obj, i)
            else:
                obj = self._env.getitem(obj, i)
        return obj, first


class SandboxedEscapeFormatter(SandboxedFormatter, EscapeFormatter):
    pass



import Cython
from cython_json import source_to_dict
import pytest
import json


def test_dump_ast_error():
    as_dict = source_to_dict(u"x = [a  10]")
    errors = as_dict['errors']
    assert len(errors) == 1
    error = errors[0]
    assert error['__node__'] == 'CompileError'
    assert error['line'] == 1
    assert error['col'] == 8
    assert 'Expected' in error['message_only']


def test_dump_error():
    contents = u'''
from distutils import sysconfig
'''
    if isinstance(contents, bytes):
        contents = contents.decode('utf-8')
    source_to_dict(contents)

def test_dump_class():
    contents = u'''
class A:pass
'''
    if isinstance(contents, bytes):
        contents = contents.decode('utf-8')
    source_to_dict(contents)

def test_comp():
    contents = u'''
{i: j for i, j in a}
'''
    if isinstance(contents, bytes):
        contents = contents.decode('utf-8')
    source_to_dict(contents)

def test_global():
    contents = u'''
def method():
  global b
  b = 10
'''
    if isinstance(contents, bytes):
        contents = contents.decode('utf-8')
    source_to_dict(contents)

# def test_dump_custom():
#     with open(r'X:\cython\tests\compile\buildenv.pyx', 'r') as stream:
#         contents = stream.read().decode('utf-8')
#     source_to_dict(contents)


def test_dump_ast():
    data = source_to_dict(u"x = [a, 10]")
    assert not data['errors']
    assert data['ast']['stats'] == [
        {
            "__node__": "SingleAssignment",
            "rhs": {
                "__node__": "List",
                "line": 1,
                "args": [
                    {
                        "__node__": "Name",
                        "line": 1,
                        "col": 5,
                        "name": "a"
                    },
                    {
                        "is_c_literal": "None",
                        "unsigned": "",
                        "value": "10",
                        "constant_result": "10",
                        "__node__": "Int",
                        "line": 1,
                        "type": "long",
                        "col": 8,
                        "longness": ""
                    }
                ],
                "col": 4
            },
            "lhs": {
                "__node__": "Name",
                "line": 1,
                "col": 0,
                "name": "x"
            },
            "line": 1,
            "col": 4
        }
    ]


if __name__ == '__main__':
    pytest.main()



from _pydevd_bundle import pydevd_timeout
from _pydevd_bundle.pydevd_constants import IS_PYPY, IS_CPYTHON
from _pydevd_bundle.pydevd_timeout import create_interrupt_this_thread_callback
from tests_python.debugger_unittest import wait_for_condition
import pytest
import threading
import time


@pytest.yield_fixture(autouse=True)
def _enable_debug_msgs():
    original = pydevd_timeout._DEBUG
    pydevd_timeout._DEBUG = True
    yield
    pydevd_timeout._DEBUG = original


def test_timeout():

    class _DummyPyDb(object):

        def __init__(self):
            self.created_pydb_daemon_threads = {}

    raise_timeout = pydevd_timeout.create_interrupt_this_thread_callback()

    def on_timeout(arg):
        assert arg == 1
        raise_timeout()

    py_db = _DummyPyDb()
    timeout_tracker = pydevd_timeout.TimeoutTracker(py_db)
    try:
        if IS_PYPY:
            timeout = 2
        else:
            timeout = 20
        with timeout_tracker.call_on_timeout(1, on_timeout, kwargs={'arg': 1}):
            time.sleep(timeout)
    except KeyboardInterrupt:
        pass


def test_timeout_0_time():

    class _DummyPyDb(object):

        def __init__(self):
            self.created_pydb_daemon_threads = {}

    raise_timeout = pydevd_timeout.create_interrupt_this_thread_callback()

    def on_timeout(arg):
        assert arg == 1
        raise_timeout()

    py_db = _DummyPyDb()
    timeout_tracker = pydevd_timeout.TimeoutTracker(py_db)
    try:
        with timeout_tracker.call_on_timeout(0, on_timeout, kwargs={'arg': 1}):
            time.sleep(1)
    except KeyboardInterrupt:
        pass


@pytest.mark.skipif(not IS_CPYTHON, reason='This only works in CPython.')
def test_create_interrupt_this_thread_callback():

    class MyThread(threading.Thread):

        def __init__(self):
            threading.Thread.__init__(self)
            self.finished = False
            self.daemon = True
            self.interrupt_thread = None
            self.interrupted = False

        def run(self):
            try:
                self.interrupt_thread = create_interrupt_this_thread_callback()
                while True:
                    time.sleep(.2)
            except KeyboardInterrupt:
                self.interrupted = True
            finally:
                self.finished = True

    t = MyThread()
    t.start()
    wait_for_condition(lambda: t.interrupt_thread is not None)

    t.interrupt_thread()

    wait_for_condition(lambda: t.finished)

    assert t.interrupted


@pytest.mark.skipif(True, reason='Skipping because running this test can interrupt the test suite execution.')
def test_interrupt_main_thread():

    from _pydevd_bundle.pydevd_constants import IS_PY39_OR_GREATER

    class MyThread(threading.Thread):

        def __init__(self, interrupt_thread_callback):
            threading.Thread.__init__(self)
            self.interrupt_thread_callback = interrupt_thread_callback

        def run(self):
            time.sleep(.5)
            self.interrupt_thread_callback()

    initial_time = time.time()
    interrupt_only_on_callback = IS_PYPY or IS_PY39_OR_GREATER
    if interrupt_only_on_callback:
        timeout = 2
    else:
        timeout = 20
    try:
        t = MyThread(create_interrupt_this_thread_callback())
        t.start()
        time.sleep(timeout)
    except KeyboardInterrupt:
        if not interrupt_only_on_callback:
            assert time.time() - initial_time < timeout
    else:
        raise AssertionError('Expected main thread to be interrupted.')



from django.shortcuts import render

# Create your views here.
from django.http import HttpResponse, HttpResponseRedirect
import sys
from .forms import NameForm


class Entry(object):

    def __init__(self, key, val):
        self.key = key
        self.val = val

    def __unicode__(self):
        return u'%s:%s' % (self.key, self.val)

    def __str__(self):
        return u'%s:%s' % (self.key, self.val)


def index(request):
    context = {
        'entries': [Entry('v1', 'v1'), Entry('v2', 'v2')]
    }
    ret = render(request, 'my_app/index.html', context)
    return ret


def get_name(request):
    # if this is a POST request we need to process the form data
    if request.method == 'POST':
        # create a form instance and populate it with data from the request:
        form = NameForm(request.POST)
        # check whether it's valid:
        if form.is_valid():
            # process the data in form.cleaned_data as required
            # ...
            # redirect to a new URL:
            return HttpResponseRedirect('/thanks/')

    # if a GET (or any other method) we'll create a blank form
    else:
        form = NameForm(data={'your_name': 'unknown name'})

    return render(request, 'my_app/name.html', {'form': form})


def inherits(request):
    context = {}
    ret = render(request, 'my_app/inherits.html', context)
    return ret


def template_error(request):
    context = {
        'entries': [Entry('v1', 'v1'), Entry('v2', 'v2')]
    }

    ret = render(request, 'my_app/template_error.html', context)
    return ret


def template_error2(request):
    context = {}
    ret = render(request, 'my_app/template_error2.html', context)
    return ret


def no_var_error(request):
    context = {}
    ret = render(request, 'my_app/no_var_error.html', context)
    return ret



from robocorp_ls_core.pluginmanager import PluginManager
import threading
from robocorp_ls_core.basic import (
    log_and_silence_errors,
    kill_process_and_subprocesses,
    is_process_alive,
)
import sys
import weakref
import os
from robocorp_ls_core.robotframework_log import get_logger
from typing import Any, Dict, Optional, Tuple, List, Iterable
from robotframework_ls.ep_resolve_interpreter import (
    EPResolveInterpreter,
    IInterpreterInfo,
)
from robocorp_ls_core.protocols import (
    IConfig,
    IMessageMatcher,
    IWorkspace,
    IRobotFrameworkApiClient,
)
import itertools
from functools import partial

DEFAULT_API_ID = "default"

log = get_logger(__name__)

_next_id = partial(next, itertools.count(0))


class _ServerApi(object):
    """
    Note: this is mainly a helper to manage the startup of an IRobotFrameworkApiClient
    and restart it when needed.

    This class is not thread-safe and should be accessed only from a single thread.

    The provided `IRobotFrameworkApiClient` may later be accessed from any thread.
    """

    def __init__(
        self,
        log_extension,
        language_server_ref,
        pre_generate_libspecs: bool = False,
        index_workspace: bool = False,
        collect_tests: bool = False,
    ) -> None:
        self._main_thread = threading.current_thread()

        from robotframework_ls.robot_config import RobotConfig

        self._used_python_executable: Optional[str] = None
        self._used_environ: Optional[Dict[str, str]] = None
        self._server_process = None

        from robotframework_ls.server_api.client import RobotFrameworkApiClient

        self._robotframework_api_client: Optional[RobotFrameworkApiClient] = None

        # We have a version of the config with the settings passed overridden
        # by the settings of a given (customized) interpreter.
        self._config: IConfig = RobotConfig()
        self._pre_generate_libspecs = pre_generate_libspecs
        self._index_workspace = index_workspace
        self._collect_tests = collect_tests

        self.workspace: Optional[IWorkspace] = None
        self._initializing = False
        self._log_extension = log_extension
        self._language_server_ref = language_server_ref
        self._interpreter_info: Optional[IInterpreterInfo] = None

        self._last_settings_sent: Optional[dict] = None

    @property
    def stats(self) -> Optional[dict]:
        client = self._robotframework_api_client
        if client is not None:
            return client.stats
        return None

    def _check_in_main_thread(self):
        curr_thread = threading.current_thread()
        if self._main_thread is not curr_thread:
            raise AssertionError(
                f"This may only be called at the thread: {self._main_thread}. Current thread: {curr_thread}"
            )

    @property
    def robot_framework_language_server(self):
        return self._language_server_ref()

    @property
    def workspace(self) -> Optional[IWorkspace]:
        return self._workspace

    @workspace.setter
    def workspace(self, workspace: IWorkspace):
        self._check_in_main_thread()
        self._workspace = workspace

    @property
    def config(self) -> IConfig:
        return self._config

    @config.setter
    def config(self, config: IConfig):
        self._check_in_main_thread()
        self._config.update(config.get_full_settings())
        self._check_reinitialize_and_forward_settings_if_needed()

    def set_interpreter_info(self, interpreter_info: IInterpreterInfo) -> None:
        from robotframework_ls.config_extension import apply_interpreter_info_to_config

        self._check_in_main_thread()
        self._interpreter_info = interpreter_info
        apply_interpreter_info_to_config(self._config, interpreter_info)

        self._check_reinitialize_and_forward_settings_if_needed()

    def _check_reinitialize_and_forward_settings_if_needed(self) -> None:
        self._check_in_main_thread()
        was_disposed = self._check_reinitialize()
        if not was_disposed:
            new_settings = self._config.get_full_settings()
            if new_settings != self._last_settings_sent:
                self._last_settings_sent = new_settings
                # i.e.: when the interpreter info changes, even if it kept the same
                # interpreter, it's possible that the configuration changed.
                self.forward(
                    "workspace/didChangeConfiguration",
                    {"settings": new_settings},
                )

    def _check_reinitialize(self) -> bool:
        """
        Returns True if the existing process was disposed (or if it wasn't even
        started) and False if the existing process was kept running.
        """
        self._check_in_main_thread()
        if self._server_process is None:
            return True

        # If the python executable changes, restart the server API.
        if self._used_python_executable is not None:
            python_executable = self._get_python_executable()
            if python_executable != self._used_python_executable:
                # It'll be reinitialized when needed.
                self._dispose_server_process()
                return True
        if self._used_environ is not None:
            environ = self._get_environ()
            if environ != self._used_environ:
                # It'll be reinitialized when needed.
                self._dispose_server_process()
                return True
        return False

    def _get_python_executable(self) -> str:
        self._check_in_main_thread()
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_PYTHON_EXECUTABLE,
        )

        config = self._config
        python_exe = sys.executable
        if config is not None:
            python_exe = config.get_setting(
                OPTION_ROBOT_PYTHON_EXECUTABLE, str, default=python_exe
            )
        else:
            log.warning(f"self._config not set in {self.__class__}")
        return python_exe

    def _get_environ(self) -> Dict[str, str]:
        self._check_in_main_thread()
        from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_PYTHON_ENV

        config = self._config
        env = os.environ.copy()

        env.pop("PYTHONPATH", "")
        env.pop("PYTHONHOME", "")
        env.pop("VIRTUAL_ENV", "")

        if config is not None:
            env_in_settings = config.get_setting(
                OPTION_ROBOT_PYTHON_ENV, dict, default={}
            )
            for key, val in env_in_settings.items():
                env[str(key)] = str(val)
        else:
            log.warning("self._config not set in %s" % (self.__class__,))
        return env

    def get_robotframework_api_client(self) -> Optional[IRobotFrameworkApiClient]:

        self._check_in_main_thread()
        workspace = self.workspace
        assert (
            workspace
        ), "The workspace must be already set when getting the server api."

        server_process = self._server_process

        if server_process is not None:
            # If someone killed it, dispose of internal references
            # and create a new process.
            if not is_process_alive(server_process.pid):
                server_process = None
                self._dispose_server_process()

        if server_process is None:
            try:
                from robotframework_ls.options import Setup
                from robotframework_ls.server_api.client import RobotFrameworkApiClient
                from robotframework_ls.server_api.server__main__ import (
                    start_server_process,
                )
                from robocorp_ls_core.jsonrpc.streams import (
                    JsonRpcStreamWriter,
                    JsonRpcStreamReader,
                )
                from robotframework_ls.robotframework_ls_impl import (
                    RobotFrameworkLanguageServer,
                )

                args = []
                if Setup.options.verbose:
                    args.append("-" + "v" * int(Setup.options.verbose))
                if Setup.options.log_file:
                    log_id = _next_id()
                    # i.e.: use a log id in case we create more than one in the
                    # same session.
                    if log_id == 0:
                        args.append(
                            "--log-file=" + Setup.options.log_file + self._log_extension
                        )
                    else:
                        args.append(
                            "--log-file="
                            + Setup.options.log_file
                            + (".%s" % (log_id,))
                            + self._log_extension
                        )

                if self._pre_generate_libspecs:
                    args.append("--pre-generate-libspecs")

                if self._index_workspace:
                    args.append("--index-workspace")

                if self._collect_tests:
                    args.append("--collect-tests")

                python_exe = self._get_python_executable()
                environ = self._get_environ()

                self._used_python_executable = python_exe
                self._used_environ = environ

                robot_framework_language_server: RobotFrameworkLanguageServer = (
                    self.robot_framework_language_server
                )
                remote_fs_observer_port = (
                    robot_framework_language_server.get_remote_fs_observer_port()
                )
                if not remote_fs_observer_port:
                    raise RuntimeError(
                        f"Expected the port to hear the Remote filesystem observer to be available. Found: {remote_fs_observer_port}"
                    )

                args.append(f"--remote-fs-observer-port={remote_fs_observer_port}")
                server_process = start_server_process(
                    args=args, python_exe=python_exe, env=environ
                )

                self._server_process = server_process

                write_to = server_process.stdin
                read_from = server_process.stdout
                w = JsonRpcStreamWriter(write_to, sort_keys=True)
                r = JsonRpcStreamReader(read_from)

                language_server_ref = self._language_server_ref

                def on_received_message(msg):
                    method = msg.get("method")

                    if method in ("$/customProgress", "$/testsCollected"):
                        robot_framework_language_server = language_server_ref()
                        if robot_framework_language_server is not None:
                            robot_framework_language_server.forward_msg(msg)

                    # Note: note done because our caches are removed promptly
                    # for this to work it should be invalidate but the info
                    # should be kept around so that we lint dependencies always
                    # not just on the first cache invalidation.
                    # WIP: test_dependency_graph_integration_lint
                    # elif method == "$/dependencyChanged":
                    #     robot_framework_language_server = language_server_ref()
                    #     if robot_framework_language_server is not None:
                    #         params = msg.get("params")
                    #         if params:
                    #             uri = params.get("uri")
                    #             if uri:
                    #                 robot_framework_language_server.lint(
                    #                     doc_uri=uri, is_saved=False
                    #                 )

                api = self._robotframework_api_client = RobotFrameworkApiClient(
                    w, r, server_process, on_received_message=on_received_message
                )

                log.debug(
                    "Initializing api... (this pid: %s, api pid: %s).",
                    os.getpid(),
                    server_process.pid,
                )
                api.initialize(
                    process_id=os.getpid(),
                    root_uri=workspace.root_uri,
                    workspace_folders=list(
                        {"uri": folder.uri, "name": folder.name}
                        for folder in workspace.iter_folders()
                    ),
                )

                config = self._config
                log.debug("Forwarding config to api...")
                if config is not None:
                    api.forward(
                        "workspace/didChangeConfiguration",
                        {"settings": config.get_full_settings()},
                    )

                # Open existing documents in the API.
                source: Optional[str]
                for document in workspace.iter_documents():
                    log.debug("Forwarding doc: %s to api...", document.uri)
                    try:
                        source = document.source
                    except Exception:
                        source = None

                    api.forward(
                        "textDocument/didOpen",
                        {
                            "textDocument": {
                                "uri": document.uri,
                                "version": document.version,
                                "text": source,
                            }
                        },
                    )

            except Exception as e:
                if server_process is None:
                    log.exception(
                        "Error starting robotframework server api (server_process=None)."
                    )
                else:
                    exitcode = server_process.poll()
                    if exitcode is not None:
                        # Note: only read() if the process exited.
                        log.exception(
                            "Error starting robotframework server api. Exit code: %s Base exception: %s. Stderr: %s",
                            exitcode,
                            e,
                            server_process.stderr.read(),
                        )
                    else:
                        log.exception(
                            "Error (%s) starting robotframework server api (still running). Base exception: %s.",
                            exitcode,
                            e,
                        )
                self._dispose_server_process()
            finally:
                if server_process is not None:
                    log.debug(
                        "Server api (%s) created pid: %s", self, server_process.pid
                    )
                else:
                    log.debug(
                        "server_process == None in get_robotframework_api_client()"
                    )

        return self._robotframework_api_client

    @log_and_silence_errors(log)
    def _dispose_server_process(self):
        self._check_in_main_thread()
        try:
            log.debug("Dispose server process.")
            if self._server_process is not None:
                if is_process_alive(self._server_process.pid):
                    kill_process_and_subprocesses(self._server_process.pid)
        finally:
            self._server_process = None
            self._robotframework_api_client = None
            self._used_environ = None
            self._used_python_executable = None
            self._last_settings_sent = None

    def request_cancel(self, message_id) -> None:
        self._check_in_main_thread()
        api = self.get_robotframework_api_client()
        if api is not None:
            api.request_cancel(message_id)

    @log_and_silence_errors(log)
    def forward(self, method_name, params) -> None:
        self._check_in_main_thread()
        api = self.get_robotframework_api_client()
        if api is not None:
            return api.forward(method_name, params)
        return None

    @log_and_silence_errors(log)
    def forward_async(self, method_name, params) -> Optional[IMessageMatcher]:
        self._check_in_main_thread()
        api = self.get_robotframework_api_client()
        if api is not None:
            return api.forward_async(method_name, params)
        return None

    @log_and_silence_errors(log)
    def open(self, uri, version, source):
        self._check_in_main_thread()
        api = self.get_robotframework_api_client()
        if api is not None:
            api.open(uri, version, source)

    @log_and_silence_errors(log)
    def exit(self):
        self._check_in_main_thread()
        if self._robotframework_api_client is not None:
            # i.e.: only exit if it was started in the first place.
            self._robotframework_api_client.exit()
        self._dispose_server_process()

    @log_and_silence_errors(log)
    def shutdown(self):
        self._check_in_main_thread()
        if self._robotframework_api_client is not None:
            # i.e.: only shutdown if it was started in the first place.
            self._robotframework_api_client.shutdown()


class _RegularLintAndOthersApi(object):
    """
    This encapsulates 3 different processes (each process is an API).

    The default (api) is usually used for requests which are real-time,
    such as code-completion, find definition, signature help and hover.

    The lint api is used for linting.

    The others api is used for other requests which are a middle ground between
    the lint (slowest) and the default (fastest). It covers requests such as
    document formatting, code folding, semantic tokens and workspace symbols.
    """

    def __init__(self, api: _ServerApi, lint_api: _ServerApi, others_api: _ServerApi):
        self.api = api
        self.lint_api = lint_api
        self.others_api = others_api

    def __iter__(self):
        yield self.api
        yield self.lint_api
        yield self.others_api

    def set_interpreter_info(self, interpreter_info: IInterpreterInfo):
        for api in self:
            api.set_interpreter_info(interpreter_info)


class ServerManager(object):
    """
    Note: accessing the ServerManager may only be done from a single thread.

    The idea is that clients do something as:

    rf_api_client = server_manager.get_lint_rf_api_client(doc_uri)
    if rf_api_client is not None:
        ... robotframework_api_client may then be accessed by any thread.
    """

    def __init__(
        self,
        pm: PluginManager,
        config: Optional[IConfig] = None,
        workspace: Optional[IWorkspace] = None,
        language_server: Optional[Any] = None,
    ):
        self._main_thread = threading.current_thread()
        self._config: Optional[IConfig] = config
        self._workspace: Optional[IWorkspace] = workspace
        self._pm = pm
        self._id_to_apis: Dict[str, _RegularLintAndOthersApi] = {}
        if language_server is None:
            self._language_server_ref = lambda: None
        else:
            self._language_server_ref = weakref.ref(language_server)

    def _check_in_main_thread(self):
        curr_thread = threading.current_thread()
        if self._main_thread is not curr_thread:
            raise AssertionError(
                f"This may only be called at the thread: {self._main_thread}. Current thread: {curr_thread}"
            )

    def _iter_all_apis(self) -> Iterable[_ServerApi]:
        self._check_in_main_thread()
        for apis in self._id_to_apis.values():
            for api in apis:
                yield api

    def set_config(self, config: IConfig) -> None:
        self._check_in_main_thread()
        self._config = config
        for api in self._iter_all_apis():
            api.config = config

    def set_workspace(self, workspace: IWorkspace) -> None:
        self._check_in_main_thread()
        self._workspace = workspace

        for api in self._iter_all_apis():
            api.workspace = workspace

    def _create_apis(self, api_id, collect_tests=False) -> _RegularLintAndOthersApi:
        self._check_in_main_thread()
        assert api_id not in self._id_to_apis, f"{api_id} already created."
        api = _ServerApi(
            ".api",
            self._language_server_ref,
            index_workspace=True,
            collect_tests=collect_tests,
        )

        lint_api = _ServerApi(
            ".lint.api", self._language_server_ref, pre_generate_libspecs=True
        )

        others_api = _ServerApi(".others.api", self._language_server_ref)

        config = self._config
        if config is not None:
            api.config = config
            lint_api.config = config
            others_api.config = config

        workspace = self._workspace
        if workspace is not None:
            api.workspace = workspace
            lint_api.workspace = workspace
            others_api.workspace = workspace

        apis = _RegularLintAndOthersApi(api, lint_api, others_api)
        self._id_to_apis[api_id] = apis
        return apis

    def _get_default_apis(self) -> _RegularLintAndOthersApi:
        self._check_in_main_thread()
        apis = self._id_to_apis.get(DEFAULT_API_ID)
        if not apis:
            apis = self._create_apis(DEFAULT_API_ID, collect_tests=True)
        return apis

    def _get_apis_for_doc_uri(self, doc_uri: str) -> _RegularLintAndOthersApi:
        self._check_in_main_thread()
        if doc_uri:
            for ep in self._pm.get_implementations(EPResolveInterpreter):
                interpreter_info = ep.get_interpreter_info_for_doc_uri(doc_uri)
                if interpreter_info is not None:
                    # Note: we currently only identify things through the interpreter
                    # id, but a potential optimization would be using the same python
                    # executable in different APIs if they match.
                    interpreter_id = interpreter_info.get_interpreter_id()
                    apis = self._id_to_apis.get(interpreter_id)
                    if apis is not None:
                        apis.set_interpreter_info(interpreter_info)
                    else:
                        apis = self._create_apis(interpreter_id)
                        apis.set_interpreter_info(interpreter_info)

                    return apis

        return self._get_default_apis()

    def forward(self, target: Tuple[str, ...], method_name: str, params: Any) -> None:
        self._check_in_main_thread()
        apis: _RegularLintAndOthersApi
        for apis in self._id_to_apis.values():
            # Note: always forward async to all APIs (all the messages are sent
            # from the current main thread, so, the messages ordering is still
            # guaranteed to be correct).
            if "api" in target:
                apis.api.forward_async(method_name, params)

            if "lint" in target:
                apis.lint_api.forward_async(method_name, params)

            if "others" in target:
                apis.others_api.forward_async(method_name, params)

    def shutdown(self) -> None:
        self._check_in_main_thread()
        for api in self._iter_all_apis():
            api.shutdown()

    def exit(self) -> None:
        self._check_in_main_thread()
        for api in self._iter_all_apis():
            api.exit()

    def collect_apis(self) -> List[_ServerApi]:
        return list(self._iter_all_apis())

    # Private APIs

    def _get_others_api(self, doc_uri: str) -> _ServerApi:
        self._check_in_main_thread()
        apis = self._get_apis_for_doc_uri(doc_uri)
        return apis.others_api

    def _get_lint_api(self, doc_uri: str) -> _ServerApi:
        self._check_in_main_thread()
        apis = self._get_apis_for_doc_uri(doc_uri)
        return apis.lint_api

    def _get_regular_api(self, doc_uri: str) -> _ServerApi:
        self._check_in_main_thread()
        apis = self._get_apis_for_doc_uri(doc_uri)
        return apis.api

    # Public APIs -- returns a client that can be accessed in any thread

    def get_lint_rf_api_client(
        self, doc_uri: str
    ) -> Optional[IRobotFrameworkApiClient]:
        """
        To be used for:

        - linting
        """
        api = self._get_lint_api(doc_uri)
        if api is not None:
            return api.get_robotframework_api_client()
        return None

    def get_regular_rf_api_client(
        self, doc_uri: str
    ) -> Optional[IRobotFrameworkApiClient]:
        """
        To be used for things that require workspace information (this is the
        one that indexes the workspace):

        i.e.:

        - auto-import completions
        - find references
        - signature help
        - collect tests
        - workspace tokens
        - find definition
        - code-completion
        """
        api = self._get_regular_api(doc_uri)
        if api is not None:
            return api.get_robotframework_api_client()
        return None

    def get_others_api_client(self, doc_uri) -> Optional[IRobotFrameworkApiClient]:
        """
        To be used for assorted things:

        - semantic tokens
        - list tests for uri
        - formatting
        - folding range
        - code lens
        - document symbols
        """
        api = self._get_others_api(doc_uri)
        if api is not None:
            return api.get_robotframework_api_client()
        return None



from threading import Thread, Event


def create_thread():
    event = Event()
    event_set = [False]

    def run_thread():
        event_set[0] = True
        event.set()
    
    t = Thread(target=run_thread)
    t.start()
    
    try:
        event.wait(5)
        
        # note: not using `assert event.wait(5)` for py2.6 compatibility.
        assert event_set[0]  
        print('TEST SUCEEDED!')
        return 'create_thread:ok'
    except:
        import traceback;traceback.print_exc()

        
a = 10  # Add breakpoint here and evaluate create_thread()



def call_method():

    from _pydevd_bundle.pydevd_constants import get_global_debugger
    py_db = get_global_debugger()

    # Check state prior to breaking
    assert not py_db.gui_in_use
    assert py_db._installed_gui_support
    assert py_db._gui_event_loop == 'qt5'

    import  os
    import PySide2
    from PySide2.QtCore import QTimer

    dirname = os.path.dirname(PySide2.__file__)
    plugin_path = os.path.join(dirname, 'plugins', 'platforms')
    if os.path.exists(plugin_path):
        os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = plugin_path

    from PySide2 import QtWidgets

    app = QtWidgets.QApplication([])

    def on_timeout():
        print('on_timeout() called')

    print_timer = QTimer()
    print_timer.timeout.connect(on_timeout)
    print_timer.setInterval(100)
    print_timer.start()

    def on_break():
        print('break here')
        app.quit()

    break_on_timer = QTimer()
    break_on_timer.timeout.connect(on_break)
    break_on_timer.setSingleShot(True)
    break_on_timer.setInterval(50)
    break_on_timer.start()

    app.exec_()  # Run forever until app.quit()

    assert py_db.gui_in_use
    assert py_db._installed_gui_support
    assert py_db._gui_event_loop == 'qt5'


if __name__ == '__main__':
    call_method()
    print('TEST SUCEEDED!')



from multiprocessing import Process, Queue


class Foo:

    def __init__(self, value):
        self.value = value  # break 2 here


def f(q):
    q.put(Foo(1))


if __name__ == '__main__':
    q = Queue()
    p = Process(target=f, args=(q,))
    p.start()
    print(q.get().value)  # break 1 here
    print('TEST SUCEEDED!')
    p.join()



from typing import Optional, Callable


class MainLoopCallbackHolder:
    ON_MAIN_LOOP: Optional[Callable[[], None]] = None


def interpreter_main_loop():
    on_main_loop = MainLoopCallbackHolder.ON_MAIN_LOOP
    assert on_main_loop
    on_main_loop()  # pylint: disable=not-callable



# this test requires numpy to be installed
import numpy

def main():
    smallarray = numpy.arange(100) * 1 + 1j
    bigarray = numpy.arange(100000).reshape((10,10000)) # 100 thousand
    hugearray = numpy.arange(10000000)  # 10 million

    pass  # location of breakpoint after all arrays defined

main()
print('TEST SUCEEDED')



def check_backtrack(x):  # line 1
    if not (x == 'a'  # line 2
        or x == 'c'):  # line 3
        pass  # line 4


import re
import sys

en_lang_symbols = r'[^\w!@#$%\^-_+=|\}{][\"\';:?\/><.,&)(*\s`\u2019]'
en_words_basic = []
en_words = []


class Dummy:
    non_en_words_limit = 3

    @staticmethod
    def fun(text):
        words = tuple(w[0].lower() for w in re.finditer(r'[a-zA-Z]+', text))
        non_en_pass = []
        for i, word in enumerate(words):
            non_en = []
            if not (word in en_words_basic
                    or (word.endswith('s') and word[:-1] in en_words_basic)
                    or (word.endswith('ed') and word[:-2] in en_words_basic)
                    or (word.endswith('ing') and word[:-3] in en_words_basic)
                    or word in en_words
                    or (word.endswith('s') and word[:-1] in en_words)
                    or (word.endswith('ed') and word[:-2] in en_words)
                    or (word.endswith('ing') and word[:-3] in en_words)
                    ):

                non_en.append(word)
                non_en_pass.append(word)
                for j in range(1, Dummy.non_en_words_limit):
                    if i + j >= len(words):
                        break
                    word = words[i + j]

                    if (word in en_words_basic
                        or (word.endswith('s') and word[:-1] in en_words_basic)
                        or (word.endswith('ed') and word[:-2] in en_words_basic)
                        or (word.endswith('ing') and word[:-3] in en_words_basic)
                        or word in en_words
                        or (word.endswith('s') and word[:-1] in en_words)
                        or (word.endswith('ed') and word[:-2] in en_words)
                        or (word.endswith('ing') and word[:-3] in en_words)
                    ):
                        break
                    else:
                        non_en.append(word)
                        non_en_pass.append(word)


def offset_overflow(stream=sys.stdout):
    a = 1
    b = 2
    c = 3
    a1 = 1 if a > 1 else 2
    a2 = 1 if a > 1 else 2
    a3 = 1 if a > 1 else 2
    a4 = 1 if a > 1 else 2
    a5 = 1 if a > 1 else 2
    a6 = 1 if a > 1 else 2
    a7 = 1 if a > 1 else 2
    a8 = 1 if a > 1 else 2
    a9 = 1 if a > 1 else 2
    a10 = 1 if a > 1 else 2
    a11 = 1 if a > 1 else 2
    a12 = 1 if a > 1 else 2
    a13 = 1 if a > 1 else 2

    for i in range(1):
        if a > 0:
            stream.write("111\n")
            # a = 1
        else:
            stream.write("222\n")
    return b


def long_lines():
    a = 1
    b = 1 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23
    c = 1 if b > 1 else 2 if b > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23
    d = 1 if c > 1 else 2 if c > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23 if a > 1 else 2 if a > 0 else 3 if a > 4 else 23
    e = d + 1
    return e



import sys
from _pydevd_bundle.pydevd_constants import IS_PY38_OR_GREATER
import pytest

SOME_LST = ["foo", "bar"]
BAR = "bar"
FOO = "foo"
global_frame = sys._getframe()


def obtain_frame():
    yield sys._getframe()


@pytest.fixture
def disable_critical_log():
    # We want to hide the logging related to _evaluate_with_timeouts not receiving the py_db.
    from _pydev_bundle.pydev_log import log_context
    import io
    stream = io.StringIO()
    with log_context(0, stream):
        yield


def test_evaluate_expression_basic(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        evaluate_expression(None, frame, 'some_var = 1', is_exec=True)

        assert frame.f_locals['some_var'] == 1

    check(next(iter(obtain_frame())))
    assert 'some_var' not in sys._getframe().f_globals

    # as locals == globals, this will also change the current globals
    check(global_frame)
    assert 'some_var' in sys._getframe().f_globals
    del sys._getframe().f_globals['some_var']
    assert 'some_var' not in sys._getframe().f_globals


def test_evaluate_expression_1(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = '''
container = ["abc","efg"]
results = []
for s in container:
    result = [s[i] for i in range(3)]
    results.append(result)
'''
        evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert frame.f_locals['results'] == [['a', 'b', 'c'], ['e', 'f', 'g']]
        assert frame.f_locals['s'] == "efg"

    check(next(iter(obtain_frame())))

    for varname in ['container', 'results', 's']:
        assert varname not in sys._getframe().f_globals

    check(global_frame)
    for varname in ['container', 'results', 's']:
        assert varname in sys._getframe().f_globals

    for varname in ['container', 'results', 's']:
        del sys._getframe().f_globals[varname]


def test_evaluate_expression_2(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = 'all((x in (BAR, FOO) for x in SOME_LST))'
        assert evaluate_expression(None, frame, eval_txt, is_exec=False)

    check(next(iter(obtain_frame())))
    check(global_frame)


def test_evaluate_expression_3(disable_critical_log):
    if not IS_PY38_OR_GREATER:
        return

    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = '''11 if (some_var := 22) else 33'''
        assert evaluate_expression(None, frame, eval_txt, is_exec=False) == 11

    check(next(iter(obtain_frame())))
    assert 'some_var' not in sys._getframe().f_globals

    # as locals == globals, this will also change the current globals
    check(global_frame)
    assert 'some_var' in sys._getframe().f_globals
    del sys._getframe().f_globals['some_var']
    assert 'some_var' not in sys._getframe().f_globals


def test_evaluate_expression_4(disable_critical_log):
    from _pydevd_bundle.pydevd_vars import evaluate_expression

    def check(frame):
        eval_txt = '''import email;email.foo_value'''
        with pytest.raises(AttributeError):
            evaluate_expression(None, frame, eval_txt, is_exec=True)
        assert 'email' in frame.f_locals

    check(next(iter(obtain_frame())))
    assert 'email' not in sys._getframe().f_globals

    # as locals == globals, this will also change the current globals
    check(global_frame)
    assert 'email' in sys._getframe().f_globals
    del sys._getframe().f_globals['email']
    assert 'email' not in sys._getframe().f_globals



def call_me_back2(callback):
    a = 'other'
    callback()
    return a


def call_me_back1(callback):
    a = 'other'
    callback()
    return a


def raise_exception():
    raise RuntimeError('TEST SUCEEDED')




import threading

semaphore1 = threading.Semaphore(0)
proceed = False


def thread_target():
    semaphore1.release()
    import time
    
    while True:
        if proceed:
            break
        time.sleep(1 / 30.)


for i in range(2):
    t = threading.Thread(target=thread_target)
    t.start()
    
semaphore1.acquire()  # let first thread run
semaphore1.acquire()  # let second thread run

# At this point we know both other threads are already running.
print('break here')

proceed = True

print('TEST SUCEEDED!')



from robot.api.parsing import ModelTransformer, EmptyLine
from robot.parsing.model.blocks import Keyword


class SmartSortKeywords(ModelTransformer):
    """
    Sort keywords in *** Keywords *** section.

    By default sorting is case insensitive, but keywords with leading underscore go to the bottom. Other underscores are
    treated as spaces.
    Empty lines (or lack of them) between keywords are preserved.

    Following code:

        *** Keywords ***
        _my secrete keyword
            Kw2

        My Keyword
            Kw1


        my_another_cool_keyword
        my another keyword
            Kw3

    Will be transformed to:

        *** Keywords ***
        my_another_cool_keyword

        my another keyword
            Kw3


        My Keyword
            Kw1
        _my secrete keyword
            Kw2

    Default behaviour could be changed using following parameters: ``case_insensitive = True``,
    ``ignore_leading_underscore = False`` and ``ignore_other_underscore = True``.

    See https://robotidy.readthedocs.io/en/latest/transformers/SmartSortKeywords.html for more examples.
    """

    ENABLED = False

    def __init__(
        self,
        case_insensitive=True,
        ignore_leading_underscore=False,
        ignore_other_underscore=True,
    ):
        self.ci = case_insensitive
        self.ilu = ignore_leading_underscore
        self.iou = ignore_other_underscore

    def visit_KeywordSection(self, node):  # noqa
        if not node.body:
            return node
        before, after = self.leave_only_keywords(node)
        empty_lines = self.pop_empty_lines(node)
        node.body.sort(key=self.sort_function)
        self.append_empty_lines(node, empty_lines)
        node.body = before + node.body + after
        return node

    @staticmethod
    def pop_empty_lines(node):
        all_empty = []
        for kw in node.body:
            kw_empty = []
            while kw.body and isinstance(kw.body[-1], EmptyLine):
                kw_empty.insert(0, kw.body.pop())
            all_empty.append(kw_empty)
        return all_empty

    @staticmethod
    def leave_only_keywords(node):
        before = []
        after = []
        while node.body and not isinstance(node.body[0], Keyword):
            before.append(node.body.pop(0))
        while node.body and not isinstance(node.body[-1], Keyword):
            after.append(node.body.pop(-1))
        return before, after

    def sort_function(self, kw):
        name = kw.name
        if self.ci:
            name = name.casefold().upper()  # to make sure that letters go before underscore
        if self.ilu:
            name = name.lstrip("_")
        if self.iou:
            index = len(name) - len(name.lstrip("_"))
            name = name[:index] + name[index:].replace("_", " ")
        return name

    @staticmethod
    def append_empty_lines(node, empty_lines):
        for kw, lines in zip(node.body, empty_lines):
            for line in lines:
                kw.body.append(line)





def generator2():
    for i in range(5):
        yield i


def generator():
    print('start')  # break here
    yield from generator2()  # step 1
    print('end')  # step 2


if __name__ == '__main__':
    for i in generator():  # generator return
        print(i)

    print('TEST SUCEEDED!')  # step 3



import threading

from _pydevd_bundle.pydevd_utils import convert_dap_log_message_to_expression
from tests_python.debug_constants import TEST_GEVENT, IS_CPYTHON
import sys
from _pydevd_bundle.pydevd_constants import IS_WINDOWS, IS_PYPY, IS_JYTHON
import pytest
import os
from _pydevd_bundle.pydevd_thread_lifecycle import pydevd_find_thread_by_id


def test_expression_to_evaluate():
    from _pydevd_bundle.pydevd_vars import _expression_to_evaluate
    assert _expression_to_evaluate(b'expr') == b'expr'
    assert _expression_to_evaluate(b'  expr') == b'expr'
    assert _expression_to_evaluate(b'for a in b:\n  foo') == b'for a in b:\n  foo'
    assert _expression_to_evaluate(b'  for a in b:\n  foo') == b'for a in b:\nfoo'
    assert _expression_to_evaluate(b'  for a in b:\nfoo') == b'  for a in b:\nfoo'
    assert _expression_to_evaluate(b'\tfor a in b:\n\t\tfoo') == b'for a in b:\n\tfoo'

    assert _expression_to_evaluate(u'  expr') == u'expr'
    assert _expression_to_evaluate(u'  for a in expr:\n  pass') == u'for a in expr:\npass'


@pytest.mark.skipif(IS_WINDOWS, reason='Brittle on Windows.')
def test_is_main_thread():
    '''
    This is now skipped due to it failing sometimes (only on Windows).

    I (fabioz) am not 100% sure on why this happens, but when this happens the initial thread for
    the tests seems to be a non main thread.

    i.e.: With an autouse fixture with a scope='session' with the code and error message below, it's
    possible to see that at even at the `conftest` import (where indent_at_import is assigned) the
    current thread is already not the main thread.

    As far as I know this seems to be an issue in how pytest-xdist is running the tests (i.e.:
    I couldn't reproduce this without running with `python -m pytest -n 0 ...`).

    -------- Code to check error / error output ----------

    from _pydevd_bundle.pydevd_utils import is_current_thread_main_thread
    import threading
    indent_at_import = threading.get_ident()

    @pytest.yield_fixture(autouse=True, scope='session')
    def check_main_thread_session(request):
        if not is_current_thread_main_thread():
            error_msg = 'Current thread does not seem to be a main thread at the start of the session. Details:\n'
            current_thread = threading.current_thread()
            error_msg += 'Current thread: %s\n' % (current_thread,)
            error_msg += 'Current thread ident: %s\n' % (current_thread.ident,)
            error_msg += 'ident at import: %s\n' % (indent_at_import,)
            error_msg += 'curr ident: %s\n' % (threading.get_ident(),)

            if hasattr(threading, 'main_thread'):
                error_msg += 'Main thread found: %s\n' % (threading.main_thread(),)
                error_msg += 'Main thread id: %s\n' % (threading.main_thread().ident,)
            else:
                error_msg += 'Current main thread not instance of: %s (%s)\n' % (
                    threading._MainThread, current_thread.__class__.__mro__,)

>           raise AssertionError(error_msg)
E           AssertionError: Current thread does not seem to be a main thread at the start of the session. Details:
E           Current thread: <_DummyThread(Dummy-2, started daemon 7072)>
E           Current thread ident: 7072
E           ident at import: 7072
E           curr ident: 7072
E           Main thread found: <_MainThread(MainThread, started 5924)>
E           Main thread id: 5924

conftest.py:67: AssertionError
    '''
    from _pydevd_bundle.pydevd_utils import is_current_thread_main_thread
    from _pydevd_bundle.pydevd_utils import dump_threads
    if not is_current_thread_main_thread():
        error_msg = 'Current thread does not seem to be a main thread. Details:\n'
        current_thread = threading.current_thread()
        error_msg += 'Current thread: %s\n' % (current_thread,)

        if hasattr(threading, 'main_thread'):
            error_msg += 'Main thread found: %s\n' % (threading.main_thread(),)
        else:
            error_msg += 'Current main thread not instance of: %s (%s)' % (
                threading._MainThread, current_thread.__class__.__mro__,)

        from io import StringIO

        stream = StringIO()
        dump_threads(stream=stream)
        error_msg += '\n\n' + stream.getvalue()
        raise AssertionError(error_msg)

    class NonMainThread(threading.Thread):

        def run(self):
            self.is_main_thread = is_current_thread_main_thread()

    non_main_thread = NonMainThread()
    non_main_thread.start()
    non_main_thread.join()
    assert not non_main_thread.is_main_thread


def test_find_thread():
    from _pydevd_bundle.pydevd_constants import get_current_thread_id
    assert pydevd_find_thread_by_id('123') is None

    assert pydevd_find_thread_by_id(
        get_current_thread_id(threading.current_thread())) is threading.current_thread()


def check_dap_log_message(log_message, expected, evaluated, eval_locals=None):
    ret = convert_dap_log_message_to_expression(log_message)
    assert ret == expected
    assert (eval(ret, eval_locals)) == evaluated
    return ret


def test_convert_dap_log_message_to_expression():
    assert check_dap_log_message(
        'a',
        "'a'",
        'a',
    )
    assert check_dap_log_message(
        'a {a}',
        "'a %s' % (a,)",
        'a value',
        {'a': 'value'}
    )
    assert check_dap_log_message(
        'a {1}',
        "'a %s' % (1,)",
        'a 1'
    )
    assert check_dap_log_message(
        'a {  }',
        "'a '",
        'a '
    )
    assert check_dap_log_message(
        'a {1} {2}',
        "'a %s %s' % (1, 2,)",
        'a 1 2',
    )
    assert check_dap_log_message(
        'a {{22:22}} {2}',
        "'a %s %s' % ({22:22}, 2,)",
        'a {22: 22} 2'
    )
    assert check_dap_log_message(
        'a {(22,33)}} {2}',
        "'a %s} %s' % ((22,33), 2,)",
        'a (22, 33)} 2'
    )

    assert check_dap_log_message(
        'a {{1: {1}}}',
        "'a %s' % ({1: {1}},)",
        'a {1: {1}}'
    )

    # Error condition.
    assert check_dap_log_message(
        'a {{22:22} {2}',
        "'Unbalanced braces in: a {{22:22} {2}'",
        'Unbalanced braces in: a {{22:22} {2}'
    )


def test_pydevd_log():
    from _pydev_bundle import pydev_log
    import io
    from _pydev_bundle.pydev_log import log_context

    stream = io.StringIO()
    with log_context(0, stream=stream):
        pydev_log.critical('always')
        pydev_log.info('never')

    assert stream.getvalue().endswith('always\n')

    stream = io.StringIO()
    with log_context(1, stream=stream):
        pydev_log.critical('always')
        pydev_log.info('this too')
        assert stream.getvalue().endswith('always\n0.00s - this too\n')

    stream = io.StringIO()
    with log_context(0, stream=stream):
        pydev_log.critical('always %s', 1)

    assert stream.getvalue().endswith('always 1\n')

    stream = io.StringIO()
    with log_context(0, stream=stream):
        pydev_log.critical('always %s %s', 1, 2)

    assert stream.getvalue().endswith('always 1 2\n')

    stream = io.StringIO()
    with log_context(0, stream=stream):
        pydev_log.critical('always %s %s', 1)

    # Even if there's an error in the formatting, don't fail, just print the message and args.
    assert stream.getvalue().endswith('always %s %s - (1,)\n')

    stream = io.StringIO()
    with log_context(0, stream=stream):
        try:
            raise RuntimeError()
        except:
            pydev_log.exception('foo')

        assert 'foo\n' in stream.getvalue()
        assert 'raise RuntimeError()' in stream.getvalue()

    stream = io.StringIO()
    with log_context(0, stream=stream):
        pydev_log.error_once('always %s %s', 1)

    # Even if there's an error in the formatting, don't fail, just print the message and args.
    assert stream.getvalue().endswith('always %s %s - (1,)\n')


def test_pydevd_logging_files(tmpdir):
    from _pydev_bundle import pydev_log
    from _pydevd_bundle.pydevd_constants import DebugInfoHolder
    import os.path
    from _pydev_bundle.pydev_log import _LoggingGlobals

    import io
    from _pydev_bundle.pydev_log import log_context

    stream = io.StringIO()
    with log_context(0, stream=stream):
        d1 = str(tmpdir.join('d1'))
        d2 = str(tmpdir.join('d2'))

        for d in (d1, d2):
            DebugInfoHolder.PYDEVD_DEBUG_FILE = os.path.join(d, 'file.txt')
            pydev_log.initialize_debug_stream(reinitialize=True)

            assert os.path.normpath(_LoggingGlobals._debug_stream_filename) == \
                os.path.normpath(os.path.join(d, 'file.%s.txt' % os.getpid()))

            assert os.path.exists(_LoggingGlobals._debug_stream_filename)

            assert pydev_log.list_log_files(DebugInfoHolder.PYDEVD_DEBUG_FILE) == [
                _LoggingGlobals._debug_stream_filename]


def _check_tracing_other_threads():
    import pydevd_tracing
    import time
    from tests_python.debugger_unittest import wait_for_condition
    try:
        import _thread
    except ImportError:
        import thread as _thread

    # This method is called in a subprocess, so, make sure we exit properly even if we somehow
    # deadlock somewhere else.
    def dump_threads_and_kill_on_timeout():
        time.sleep(10)
        from _pydevd_bundle import pydevd_utils
        pydevd_utils.dump_threads()
        time.sleep(1)
        import os
        os._exit(77)

    _thread.start_new_thread(dump_threads_and_kill_on_timeout, ())

    def method():
        while True:
            trace_func = sys.gettrace()
            if trace_func:
                threading.current_thread().trace_func = trace_func
                break
            time.sleep(.01)

    def dummy_thread_method():
        threads.append(threading.current_thread())
        method()

    threads = []
    threads.append(threading.Thread(target=method))
    threads[-1].daemon = True
    threads[-1].start()
    _thread.start_new_thread(dummy_thread_method, ())

    wait_for_condition(lambda: len(threads) == 2, msg=lambda:'Found threads: %s' % (threads,))

    def tracing_func(frame, event, args):
        return tracing_func

    assert pydevd_tracing.set_trace_to_threads(tracing_func) == 0

    def check_threads_tracing_func():
        for t in threads:
            if getattr(t, 'trace_func', None) != tracing_func:
                return False
        return True

    wait_for_condition(check_threads_tracing_func)

    assert tracing_func == sys.gettrace()


def _build_launch_env():
    import os
    import pydevd

    environ = os.environ.copy()
    cwd = os.path.abspath(os.path.dirname(__file__))
    assert os.path.isdir(cwd)

    resources_dir = os.path.join(os.path.dirname(pydevd.__file__), 'tests_python', 'resources')
    assert os.path.isdir(resources_dir)

    attach_to_process_dir = os.path.join(os.path.dirname(pydevd.__file__), 'pydevd_attach_to_process')
    assert os.path.isdir(attach_to_process_dir)

    pydevd_dir = os.path.dirname(pydevd.__file__)
    assert os.path.isdir(pydevd_dir)

    environ['PYTHONPATH'] = (
            cwd + os.pathsep +
            resources_dir + os.pathsep +
            attach_to_process_dir + os.pathsep +
            pydevd_dir + os.pathsep +
            environ.get('PYTHONPATH', '')
    )
    return cwd, environ


def _check_in_separate_process(method_name, module_name='test_utilities', update_env={}):
    import subprocess
    cwd, environ = _build_launch_env()
    environ.update(update_env)

    subprocess.check_call(
        [sys.executable, '-c', 'import %(module_name)s;%(module_name)s.%(method_name)s()' % dict(
            method_name=method_name, module_name=module_name)],
        env=environ,
        cwd=cwd
    )


@pytest.mark.skipif(not IS_CPYTHON, reason='Functionality to trace other threads requires CPython.')
def test_tracing_other_threads():
    # Note: run this test in a separate process so that it doesn't mess with any current tracing
    # in our current process.
    _check_in_separate_process('_check_tracing_other_threads')


def _check_basic_tracing():
    import pydevd_tracing

    # Note: run this test in a separate process so that it doesn't mess with any current tracing
    # in our current process.
    called = [0]

    def tracing_func(frame, event, args):
        called[0] = called[0] + 1
        return tracing_func

    assert pydevd_tracing.set_trace_to_threads(tracing_func) == 0

    def foo():
        pass

    foo()
    assert called[0] > 2


@pytest.mark.skipif(not IS_CPYTHON, reason='Functionality to trace other threads requires CPython.')
def test_tracing_basic():
    _check_in_separate_process('_check_basic_tracing')


@pytest.mark.skipif(not IS_CPYTHON, reason='Functionality to trace other threads requires CPython.')
def test_find_main_thread_id():
    # Note: run the checks below in a separate process because they rely heavily on what's available
    # in the env (such as threads or having threading imported).
    _check_in_separate_process('check_main_thread_id_simple', '_pydevd_test_find_main_thread_id')
    _check_in_separate_process('check_main_thread_id_multiple_threads', '_pydevd_test_find_main_thread_id')
    _check_in_separate_process('check_win_threads', '_pydevd_test_find_main_thread_id')
    _check_in_separate_process('check_fix_main_thread_id_multiple_threads', '_pydevd_test_find_main_thread_id')

    import subprocess
    import pydevd
    cwd, environ = _build_launch_env()

    subprocess.check_call(
        [sys.executable, '-m', '_pydevd_test_find_main_thread_id'],
        env=environ,
        cwd=cwd
    )

    resources_dir = os.path.join(os.path.dirname(pydevd.__file__), 'tests_python', 'resources')

    subprocess.check_call(
        [sys.executable, os.path.join(resources_dir, '_pydevd_test_find_main_thread_id.py') ],
        env=environ,
        cwd=cwd
    )


@pytest.mark.skipif(not IS_WINDOWS or IS_JYTHON, reason='Windows-only test.')
def test_get_ppid():
    from _pydevd_bundle.pydevd_api import PyDevdAPI
    api = PyDevdAPI()
    # On python 3 we can check that our internal api which is used for Python 2 gives the
    # same result as os.getppid.
    ppid = os.getppid()
    assert api._get_windows_ppid() == ppid


def _check_gevent(expect_msg):
    from _pydevd_bundle.pydevd_utils import notify_about_gevent_if_needed
    assert not notify_about_gevent_if_needed()
    import gevent
    assert not notify_about_gevent_if_needed()
    import gevent.monkey
    assert not notify_about_gevent_if_needed()
    gevent.monkey.patch_all()
    assert notify_about_gevent_if_needed() == expect_msg


def check_notify_on_gevent_loaded():
    _check_gevent(True)


def check_dont_notify_on_gevent_loaded():
    _check_gevent(False)


@pytest.mark.skipif(not TEST_GEVENT, reason='Gevent not installed.')
def test_gevent_notify():
    _check_in_separate_process('check_notify_on_gevent_loaded', update_env={'GEVENT_SUPPORT': ''})
    _check_in_separate_process('check_dont_notify_on_gevent_loaded', update_env={'GEVENT_SUPPORT': 'True'})


@pytest.mark.skipif(True, reason='Skipping because running this test can interrupt the test suite execution.')
def test_interrupt_main_thread():
    from _pydevd_bundle.pydevd_utils import interrupt_main_thread
    import time

    main_thread = threading.current_thread()

    def interrupt():
        # sleep here so that the main thread in the test can get to the sleep too (otherwise
        # if we interrupt too fast we won't really check that the sleep itself
        # got interrupted -- although if that happens on some tests runs it's
        # not really an issue either).
        time.sleep(1)
        interrupt_main_thread(main_thread)

    if IS_PYPY:
        # On PyPy a time.sleep() is not being properly interrupted,
        # so, let's just check that it throws the KeyboardInterrupt in the
        # next instruction.
        timeout = 2
    else:
        timeout = 20
    initial_time = time.time()
    try:
        t = threading.Thread(target=interrupt)
        t.start()
        time.sleep(timeout)
    except KeyboardInterrupt:
        if not IS_PYPY:
            actual_timeout = time.time() - initial_time
            # If this fails it means that although we interrupted Python actually waited for the next
            # instruction to send the event and didn't really interrupt the thread.
            assert actual_timeout < timeout, 'Expected the actual timeout (%s) to be < than the timeout (%s)' % (
                actual_timeout, timeout)
    else:
        raise AssertionError('KeyboardInterrupt not generated in main thread.')


@pytest.mark.skipif(sys.version_info[0] < 3, reason='Only available for Python 3.')
def test_get_smart_step_into_variant_from_frame_offset():
    from _pydevd_bundle.pydevd_bytecode_utils import get_smart_step_into_variant_from_frame_offset
    variants = []
    assert get_smart_step_into_variant_from_frame_offset(0, variants) is None

    class DummyVariant(object):

        def __init__(self, offset):
            self.offset = offset

    variants.append(DummyVariant(1))
    assert get_smart_step_into_variant_from_frame_offset(0, variants) is None
    assert get_smart_step_into_variant_from_frame_offset(1, variants) is variants[0]
    assert get_smart_step_into_variant_from_frame_offset(2, variants) is variants[0]

    variants.append(DummyVariant(3))
    assert get_smart_step_into_variant_from_frame_offset(0, variants) is None
    assert get_smart_step_into_variant_from_frame_offset(1, variants) is variants[0]
    assert get_smart_step_into_variant_from_frame_offset(2, variants) is variants[0]
    assert get_smart_step_into_variant_from_frame_offset(3, variants) is variants[1]
    assert get_smart_step_into_variant_from_frame_offset(4, variants) is variants[1]


def test_threading_hide_pydevd():

    class T(threading.Thread):

        def __init__(self, is_pydev_daemon_thread):
            from _pydevd_bundle.pydevd_daemon_thread import mark_as_pydevd_daemon_thread
            threading.Thread.__init__(self)
            if is_pydev_daemon_thread:
                mark_as_pydevd_daemon_thread(self)
            else:
                self.daemon = True
            self.event = threading.Event()

        def run(self):
            self.event.wait(10)

    current_count = threading.active_count()
    t0 = T(True)
    t1 = T(False)
    t0.start()
    t1.start()

    # i.e.: the patching doesn't work for other implementations.
    if IS_CPYTHON:
        assert threading.active_count() == current_count + 1
        assert t0 not in threading.enumerate()
    else:
        assert threading.active_count() == current_count + 2
        assert t0 in threading.enumerate()

    assert t1 in threading.enumerate()
    t0.event.set()
    t1.event.set()


def test_import_token_from_module():
    from _pydevd_bundle.pydevd_utils import import_attr_from_module

    with pytest.raises(ImportError):
        import_attr_from_module('sys')

    with pytest.raises(ImportError):
        import_attr_from_module('sys.settrace.foo')

    assert import_attr_from_module('sys.settrace') == sys.settrace
    assert import_attr_from_module('threading.Thread.start') == threading.Thread.start



import time

if __name__ == '__main__':
    wait = True

    while wait:
        time.sleep(1)  # break here

    # Ok, if it got here things are looking good, let's just make
    # sure that the threading module main thread has the correct ident.
    import threading  # Note: only import after the attach.
    if hasattr(threading, 'main_thread'):
        assert threading.current_thread().ident == threading.main_thread().ident
    else:
        # Python 2 does not have main_thread, but we can still get the reference.
        assert threading.current_thread().ident == threading._shutdown.im_self.ident
    print('TEST SUCEEDED')



import re
from typing import Optional
import string

from robot.api.parsing import ModelTransformer, Token, KeywordCall

from robotidy.decorators import check_start_end_line
from robotidy.exceptions import InvalidParameterValueError


class RenameKeywords(ModelTransformer):
    """
    Enforce keyword naming.

    Title Case is applied to keyword name and underscores are replaced by spaces.

    You can keep underscores if you set remove_underscores to False:

        robotidy --transform RenameKeywords -c RenameKeywords:remove_underscores=False .

    It is also possible to configure `replace_pattern` parameter to find and replace regex pattern. Use `replace_to`
    to set replacement value. This configuration (underscores are used instead of spaces):

        robotidy --transform RenameKeywords -c RenameKeywords:replace_pattern=^(?i)rename\s?me$:replace_to=New_Shining_Name .

    will transform following code:

        *** Keywords ***
        rename Me
           Keyword Call

    To:

        *** Keywords ***
        New Shining Name
            Keyword Call

    Use `ignore_library = True` parameter to control if the library name part (Library.Keyword) of keyword call
    should be renamed.

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/RenameKeywords.html for more examples.
    """

    ENABLED = False

    def __init__(
        self,
        replace_pattern: Optional[str] = None,
        replace_to: Optional[str] = None,
        remove_underscores: bool = True,
        ignore_library: bool = True,
    ):
        self.ignore_library = ignore_library
        self.remove_underscores = remove_underscores
        try:
            self.replace_pattern = re.compile(replace_pattern) if replace_pattern is not None else None
        except re.error as err:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "replace_pattern",
                replace_pattern,
                f"It should be a valid regex expression. Regex error: '{err.msg}'",
            )
        self.replace_to = "" if replace_to is None else replace_to

    @check_start_end_line
    def rename_node(self, node, type_of_name):
        token = node.get_token(type_of_name)
        if not token or not token.value:
            return node
        values = []
        split_names = token.value.split(".")
        for index, value in enumerate(split_names, start=1):
            if self.ignore_library and index != len(split_names):
                values.append(value)
                continue
            if self.replace_pattern is not None:
                value = self.replace_pattern.sub(repl=self.replace_to, string=value)
            if self.remove_underscores and set(value) != {"_"}:
                value = re.sub("_+", " ", value)  # replace one or more _ with one space
            value = "".join([a if a.isupper() else b for a, b in zip(value, string.capwords(value.strip()))])
            values.append(value)
        token.value = ".".join(values)
        return node

    def visit_KeywordName(self, node):  # noqa
        return self.rename_node(node, Token.KEYWORD_NAME)

    def visit_KeywordCall(self, node):  # noqa
        return self.rename_node(node, Token.KEYWORD)



from itertools import takewhile

from robot.api.parsing import ModelTransformer, Token

try:
    from robot.api.parsing import InlineIfHeader
except ImportError:
    InlineIfHeader = None

from robotidy.decorators import check_start_end_line
from robotidy.exceptions import InvalidParameterValueError


class NormalizeSeparators(ModelTransformer):
    """
    Normalize separators and indents.

    All separators (pipes included) are converted to fixed length of 4 spaces (configurable via global argument
    ``--spacecount``).

    You can decide which sections should be transformed by configuring
    ``sections = comments,settings,variables,keywords,testcases`` param.

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/NormalizeSeparators.html for more examples.
    """

    def __init__(self, sections: str = None):
        self.indent = 0
        self.sections = self.parse_sections(sections)
        self.is_inline = False

    def parse_sections(self, sections):
        default = {"comments", "settings", "testcases", "keywords", "variables"}
        if sections is None:
            return default
        if not sections:
            return {}
        parts = sections.split(",")
        parsed_sections = set()
        for part in parts:
            part = part.replace("_", "")
            if part and part[-1] != "s":
                part += "s"
            if part not in default:
                raise InvalidParameterValueError(
                    self.__class__.__name__,
                    "sections",
                    sections,
                    f"Sections to be transformed should be provided in comma separated "
                    f"list with valid section names:\n{sorted(default)}",
                )
            parsed_sections.add(part)
        return parsed_sections

    def visit_File(self, node):  # noqa
        self.indent = 0
        return self.generic_visit(node)

    def should_visit(self, name, node):
        if name in self.sections:
            return self.generic_visit(node)
        return node

    def visit_CommentSection(self, node):  # noqa
        return self.should_visit("comments", node)

    def visit_SettingSection(self, node):  # noqa
        return self.should_visit("settings", node)

    def visit_VariableSection(self, node):  # noqa
        return self.should_visit("variables", node)

    def visit_KeywordSection(self, node):  # noqa
        return self.should_visit("keywords", node)

    def visit_TestCaseSection(self, node):  # noqa
        return self.should_visit("testcases", node)

    def indented_block(self, node):
        self.visit_Statement(node.header)
        self.indent += 1
        node.body = [self.visit(item) for item in node.body]
        self.indent -= 1
        return node

    def visit_TestCase(self, node):  # noqa
        return self.indented_block(node)

    visit_Keyword = visit_While = visit_TestCase  # noqa

    def visit_For(self, node):
        node = self.indented_block(node)
        self.visit_Statement(node.end)
        return node

    def visit_Try(self, node):
        node = self.indented_block(node)
        if node.next:
            self.visit(node.next)
        if node.end:
            self.visit_Statement(node.end)
        return node

    def visit_If(self, node):
        if self.is_inline:  # nested inline if is ignored
            return node
        self.is_inline = InlineIfHeader and isinstance(node.header, InlineIfHeader)
        self.visit_Statement(node.header)
        indent = 1
        if self.is_inline:
            indent = self.indent
            self.indent = 1
        else:
            self.indent += 1
        node.body = [self.visit(item) for item in node.body]
        if self.is_inline:
            self.indent = indent
        else:
            self.indent -= 1
        if node.orelse:
            self.visit(node.orelse)
        if node.end:
            self.visit_Statement(node.end)
        self.is_inline = False
        return node

    @check_start_end_line
    def visit_Statement(self, statement):  # noqa
        has_pipes = statement.tokens[0].value.startswith("|")
        return self._handle_spaces(statement, has_pipes)

    def _handle_spaces(self, statement, has_pipes):
        new_tokens = []
        for line in statement.lines:
            prev_sep = False
            for index, token in enumerate(line):
                if token.type == Token.SEPARATOR:
                    if prev_sep:
                        continue
                    prev_sep = True
                    count = self.indent if index == 0 else 1
                    token.value = self.formatting_config.separator * count
                else:
                    prev_sep = False
                if has_pipes and index == len(line) - 2:
                    token.value = token.value.rstrip()
                new_tokens.append(token)
        statement.tokens = new_tokens
        self.generic_visit(statement)
        return statement



import time
def Call():
    loop = True # Break here

    count = 0
    while(loop and (count < 200)):
        time.sleep(0.1)
        count += 1 # Pause here and change loop to False


if __name__ == '__main__':
    Call()
    print('TEST SUCEEDED!')